

#ifndef AMRLOAD_C
#define AMRLOAD_C

#define READ_BLOCK 512 // Must be bigger than 1 frame (12+1 B)
#define MSEC * 1000000ULL
#define DISCARD_TIMEOUT (5000 MSEC)
// commented: -> don't discard data
//#define DISCARD_OLD_AMR_DATA 

#include <stdbool.h>
#include <stdint.h>
#include <stdio.h> // FILE, later fopen, fread, fprintf

#define u8 uint8_t 
#define u16 uint16_t
#define u32 uint32_t
#define u64 uint64_t

#ifndef NOCODE
#define NOCODE
#include "multihw.c"
#undef NOCODE
#else
#include "multihw.c"
#endif


#if (defined request_DEBUG_ALGO)        && !(defined DEBUG_ALGO)
#define DEBUG_ALGO
#endif
#if (defined request_DEBUG_ALGO_DETAIL) && !(defined DEBUG_ALGO_DETAIL)
#define DEBUG_ALGO_DETAIL
#endif

struct amrload {
	
	// For AMR data
	FILE * source;
	u8 data[2*READ_BLOCK]; // ringbuffer
	u16 read; // read pointer
	bool write_high; // true if next write is to the high bytes
	
	
	
	// For application fine control
	u64 starting_time; // Has to be set BY YOU!
	// starting_time is only needed for time based reads.
	// time is always in nanoseconds.
	u32 cnt;
	
};

struct mr475_frame {
	u8 data[12];
};


#ifndef NOCODE
#include <errno.h>
#include <string.h>
#include "pindef.h" // AMRLOAD_FILE
#endif


// First functions only specific to AMR.


#ifndef NOCODE
bool load_next(struct amrload* al)
{
	if (!al->source)
		return false;
	
	
	if (al->write_high) {
		if (al->read >= READ_BLOCK)
			return true;
		
		al->write_high = false;
#ifdef DEBUG_FOR_VALGRIND
		printf("Loading high.\n");
#endif
		size_t read = fread(al->data + READ_BLOCK, READ_BLOCK, 1, al->source);
		return read == 1;
	}
	else {
		if (al->read < READ_BLOCK)
			return true;
		
		al->write_high = true;
#ifdef DEBUG_FOR_VALGRIND
		printf("Loading low.\n");
#endif
		size_t read = fread(al->data, READ_BLOCK, 1, al->source);
		return read == 1;
	}
	
	// TODO: There's no file end mechanism as input file will have more
	// than two hours of sound data. (Simulation only takes two hours)
}
;
#endif

bool read_amr_frame(struct amrload* al, struct mr475_frame * dst)
#ifndef NOCODE
{
	
	if (al->read == 2*READ_BLOCK)
		al->read = 0;
	
	//fprintf(stderr, "al->read: %d.\n", al->read);
	// trailing byte
	if (al->data[al->read++] != 0x04) {
		//printf("data: 0x%02x\n", al->data[al->read - 1]);
		printf("get_amr_frame(): ERROR! No data loaded!\n");
		//al->read--;
		return false;
	}
	
	if (al->read == 2*READ_BLOCK)
		al->read = 0;
	
	
	u8 cnt = 12;
	u8 second = 0;
	
	if (al->read + cnt > 2*READ_BLOCK) {
		second = cnt - (2*READ_BLOCK - al->read);
		cnt = 2*READ_BLOCK - al->read;
	}
	
	//fprintf(stderr, "al->read = %d.\n", al->read);
	//for (u16 i = 0; i < cnt; i++)
	//	fprintf(stderr, "d [%d] = 0x%02x\n", i, al->data[al->read + i]); 
	memcpy(dst->data, &(al->data[al->read]), cnt);
	
	if (second)
		memcpy(&(dst->data[cnt]), al->data, second);
	//for (u16 i = 0; second && i < second; i++)
	//	fprintf(stderr, "d[%d] = 0x%02x\n", i+cnt, al->data[i+cnt]); 
	
	al->read += 12;
	al->read %= 2*READ_BLOCK;
	//fprintf(stderr, "al->read -> %d.\n", al->read);
	
	if(!load_next(al)) {
		printf("get_amr_frame(): ERROR: Couldn't read file!\n");
		return false; // Of course it could be true as there _is_ actual data
	}
	
	return true;
}
#endif
;


bool noinit_amrload(struct amrload* al)
#ifndef NOCODE
{
	al->source = NULL;
	return true;
}
#endif
;

bool init_amrload(struct amrload* al)
#ifndef NOCODE
{
	bool res = true;
	al->source = fopen(AMRLOAD_FILE, "r");
	if (!al->source) {
		fprintf(stderr, "amrload.c:init_amrload(): ERROR %d when opening file %s: %s\n", errno, AMRLOAD_FILE, strerror(errno));
		res = false;
	}
	
	memset(al->data, 0xAB, READ_BLOCK); // This way we detect failed reads
	memset(al->data + READ_BLOCK, 0xBA, READ_BLOCK); //  when debugging.
	
	al->write_high = false;
	
	if (res) {
		al->read = READ_BLOCK; // forces load_next to actually load.
		res = load_next(al);
	}
	al->read = 6; // strlen("#!AMR\n") = 6
	
	al->cnt = 0;
	
	if (!res) {
		fprintf(stderr, "amrload.c:init_amrload(): ERROR %d when opening or reading file %s: %s\n", errno, AMRLOAD_FILE, strerror(errno));
	}
	return res;
}
#endif
;



#ifndef NO_REGULATED_AMR

// Next functions more bound to application

// For application: based on time
// time is in nanoseconds
// returns false if no frame is available
// TODO: This could ge (much) more intelligent, e. g. skip the frames which are
//   too much in the past.
bool get_amr_frame(MHI_ARG    struct amrload* al, struct mr475_frame * dst)
#ifndef NOCODE
{
	if (!al->source) return false;
	
	//printf("al->starting_time=%'lu", al->starting_time);
	// 10 SEC: Magic Number TODO
	uint64_t actual_time = get_systime(MHI_USE_single) - 10 SEC;
	
	// ceil.
	u32 req_cnt = (actual_time - al->starting_time) / 20000000;
	if (al->cnt >= req_cnt) 
		return false; // Data not yet available
	
#ifdef DISCARD_OLD_AMR_DATA
	u32 dis_cnt = (actual_time - DISCARD_TIMEOUT - al->starting_time) /20000000;
	if (actual_time < DISCARD_TIMEOUT)
		dis_cnt = 0;
	
	
#ifdef DEBUG_ALGO
	bool discard = false;
	if (al->cnt < dis_cnt)
		discard = true;
#endif
	
	while (al->cnt < dis_cnt) {
#ifdef DEBUG_ALGO
		printf(".");
#endif
		read_amr_frame(al, dst);
		al->cnt++;
	}
#ifdef DEBUG_ALGO
	if (discard)
		printf(" amrload: WARNING: DISCARDING AMR DATA!\n");
#endif
	
#endif // -ifdef DISCARD_OLD_AMR_DATA
	
	al->cnt++;
	
	
#ifdef PRINT_AMR_COUNT
	ulog_str_num_BASE_str_L("amrload: this is frame ", al->cnt, DEC, " .\n");
#endif
	return read_amr_frame(al, dst);
}
#endif // -ifndef NOCODE
;


#endif // -ifndef NO_REGULATED_AMR








#endif // -ifndef AMRLOAD_C


