#!/bin/bash

progname="$0"

usage() {
	printf 'usage: %s <elffile> <varname> <old value> <new value> <hexfile>\n' "$progname"
	printf '  where elffile contains the position information\n' 
	printf '  where varname is the variable in .data whose value is to be replaced\n'
	printf '  where old value (in hex) is the expected first byte of the variable (this is\n'
	printf '        a safety check). Check can be omitted by providing a empty string ("").\n'
	printf '  where new value is exactly one byte (in hex) which will be placed as the first\n'
	printf '        byte of the variable\n'
	printf '  where hexfile is the existing file which will be modified\n'
	#printf 'try this:  ./implant.sh built/OnlyReceive.ino.elf global_num 07 E7 built/OnlyReceive.ino.hex' 
}

if [ $# -lt 5 ]; then
	usage;
	false; exit;
fi

elffile="$1"
varname="$2"
safetyC="`tr '[a-f]' '[A-F]' <<< "$3"`"
replace="`tr '[a-f]' '[A-F]' <<< "$4"`"
hexfile="$5"

# Implants the station name into the Hex file.

check_elf() {
	beginning=`file -b "$1"`
	beginning2="${beginning#ELF 32-bit LSB executable, Atmel AVR 8-bit, }"
	if [ _"$beginning" = _"$beginning2" ]; then
		printf 'ERROR: This is not an elf file: %s\n';
		usage; false; exit
	fi
}


check_ascii() {
	beginning=`file -b "$1"`
	#beginning2="${beginning#ASCII text, with CRLF line terminators}"
	beginning2="${beginning#ASCII text}"
	if [ _"$beginning" = _"$beginning2" ]; then
		printf 'ERROR: This is not a hex file: %s\n';
		usage; false; exit
	fi	
}



# Find the data section
find_data_LMA() {
	objdump -h "$elffile" \
	| grep .data \
	| awk '{print $5}' \
	| tr '[a-f]' '[A-F]' \
	| sed -r 's#^0000##'
}

find_data_VMA() {
	objdump -h "$elffile" \
	| grep .data \
	| awk '{print $4}' \
	| tr '[a-f]' '[A-F]'
}


varaddress() {
	out=`objdump -t "$elffile" | grep -E " $varname\$"`
	if [ -z "$out" ]; then
		printf >&2 'ERROR: Couldn'"'"'t find any occurence of "%s"!\n' "$varname"
		false; exit;
	elif [ `wc -l <<< "$out"` -ne 1 ]; then
		printf >&2 'ERROR: Couldn'"'"'t find one unique occurence of "%s"!\n' "$varname"
	#	printf 'ERROR: Couldn\'t find one unique occurence of "%s"!\n' "$varname"
		printf >&2 '       (objdump output was following:)\n'
		cat >&2 <<< "$out"
		false; exit;
	fi
	awk '{print $1}' <<< "$out" | tr '[a-f]' '[A-F]'
}

checksum() {
	(
		echo -n "$hex 100-(";
		dos2unix \
			| sed -r 's#^:##;s#[0-9A-F]{2}$##;s#([0-9A-F]{2})#\1+#g' \
			| tr '\n' '0'
		echo ")%100"
	) \
	| bc \
	| sed -r 's#^([0-9A-F])$#0\1#'
}


check_elf "$elffile"
check_ascii "$hexfile"

datalma=`find_data_LMA`; [ $? -ne 0 ] && exit
datavma=`find_data_VMA`; [ $? -ne 0 ] && exit
varaddr=`varaddress`; [ $? -ne 0 ] && exit

echo $datalma
echo $datavma
echo $varaddr
echo


hex="obase=16;ibase=16;"

datasub=`bc <<< "$hex$varaddr-$datavma"`
echo $datasub
echo

subblck=`bc <<< "$hex      $datasub/10*10"`
subaddr=`bc <<< "ibase=16; $datasub%10"`

echo subblck "(hex):" $subblck
echo subaddr "(dec):" $subaddr
echo

tgtline=`bc <<< "$hex$datalma+$subblck"`
echo tgtline $tgtline
echo

tgtposl="`grep -Eob '^:[0-9A-F]{2}'${tgtline}'00' "$hexfile"`"
tgtposl=${tgtposl%%:*}

echo tgtposl $tgtposl;


# Now moving to hex file offsets
tgtposs=`dd bs=1 if="$hexfile" count=2 skip="$(($tgtposl+1))" 2>/dev/null`
tgtposs=`bc <<< "ibase=16; $tgtposs"`

echo tgtposs $tgtposs;
echo

# Pre safety check
if [ -n "`sed -r 's#[0-9]##g' <<< "$tgtposs"`" ]; then
	printf "ERROR: Cann't read line size, something is wrong. Quitting.\n"
	false; exit
fi

# Safety check
readold="`dd bs=1 if="$hexfile" count=2 skip=$(($tgtposl+9+2*$subaddr)) 2>/dev/null`"

if [ -n "$safetyC" -a \! _"$readold" = _"$safetyC" ]; then
	printf "ERROR: Safety check was requested but failed. You supposed us to "
	printf "find\n value=%s but it actually is %s.\n" "$safetyC" "$readold"
	false; exit
fi



printf "Now IMPLANTING new value...\n"
dd bs=1 of="$hexfile" count=2 seek=$(($tgtposl+9+2*$subaddr)) status=none conv=notrunc <<< "$replace"
if [ \! $? = 0 ]; then
	printf 'ERROR: Main implantation failed.\n'
	false; exit
fi
printf " ... done.\n\n"

# Adjusting Checksum
#echo line length: $((9+2*$tgtposs+2))
newline="`dd bs=1 if="$hexfile" skip=$tgtposl count=$((9+2*$tgtposs+2)) status=none`"
#newline="`grep -E '^:[0-9A-F]{2}'${tgtline}'00' "$hexfile"`" # Alternatively
#echo newline "$newline"

chksumN=`checksum <<< "$newline"`

echo chksumN $chksumN
echo

printf "Now adjusting checksum...\n"
dd bs=1 of="$hexfile" count=2 seek=$(($tgtposl+9+2*$tgtposs)) status=none conv=notrunc <<< "$chksumN"
printf " ... done.\n\n"

true; exit

