/*
 *  Interrupt and PWM utilities for 16 bit Timer1 on ATmega168/328
 *  Original code by Jesse Tane for http://labs.ideo.com August 2008
 *  Modified March 2009 by Jérôme Despatis and Jesse Tane for ATmega328 support
 *  Modified June 2009 by Michael Polli and Jesse Tane to fix a bug in set_period() which caused the timer to stop
 *
 *  This is free software. You can redistribute it and/or modify it under
 *  the terms of Creative Commons Attribution 3.0 United States License. 
 *  To view a copy of this license, visit http://creativecommons.org/licenses/by/3.0/us/ 
 *  or send a letter to Creative Commons, 171 Second Street, Suite 300, San Francisco, California, 94105, USA.
 *
 */

#include "TimerOne.h"
#include "cassert.h"

TimerOne Timer1;              // preinstatiate

ISR(TIMER1_OVF_vect)          // interrupt service routine that wraps a user defined function supplied by attach_interrupt
{
	// Set Mode to normal -> Timer only counting up
	TCCR1A  = 0;
	TCCR1B &= ~_BV(WGM13) & ~_BV(WGM12);
	
	Timer1.detachInterrupt();
	
	Timer1.isrCallback();
}

#define inline // XXX HACK short-time simply to test; inline would be better.
inline void TimerOne::initialize(long microseconds) {
	
//	// Set mode 8 (PWM, Phase and Frequency Correct, WGM13:WGM10 = 1000b).
//	// This will also stop the timers (CS1x = 0)
//	TCCR1A = 0;           // WGM11, WGM10
//	TCCR1B = _BV(WGM13);  // WGM13, WGM12, CS12, CS11, CS10
//	
//	// In this mode TCNT1 will run forward AND THEN BACKWARD. When it reaches
//	// zero, a TOV1 interrupt will be generated.
	
	// Set mode 14 (Fast PWM with ICR1, WGM13:WGM10 = 1110b).
	// This will also stop the timers (CS1x = 0)
	TCCR1A = _BV(WGM11);               // WGM11, WGM10
	TCCR1B = _BV(WGM13) | _BV(WGM12);  // WGM13, WGM12, CS12, CS11, CS10
	// In this mode TCNT1 will run forward only. When it reaches TOP (=ICR1)
	// a TOV1 interrupt will be generated and TCNT1 will be reset to 0.
	
	
	// Disable all interrupts
	//TIMSK1 &= ~(_BV(TICIE1) | _BV(OCIE1A) | _BV(OCIE1B) | _BV(TOIE1));
	TIMSK1 &= ~(_BV(OCIE1A) | _BV(OCIE1B) | _BV(TOIE1)); // doesn't find TICIE1??
	
	set_period(microseconds);
}

void TimerOne::calc_period(long microseconds) {
	
	
//	// the counter runs backwards after TOP, interrupt is
//	//at BOTTOM so divide microseconds by 2
//	long cycles = (F_CPU * microseconds) / 2000000;
	
	// mode 14
	long cycles = (F_CPU * microseconds) / 1000000;
	
	// If it was to be made efficient, the highest high bit of cycles 
	// could be determined, which is more efficient than these subtractions.
	if(cycles < RESOLUTION) {
		clockSelectBits = _BV(CS10);              // no prescale, full xtal
		prescale_factor = 0;
	}
	else if((cycles >>= 3) < RESOLUTION) {
		clockSelectBits = _BV(CS11);              // prescale by /8
		prescale_factor = 8;
	}
	else if((cycles >>= 3) < RESOLUTION) {
		clockSelectBits = _BV(CS11) | _BV(CS10);  // prescale by /64
		prescale_factor = 64;
	}
	else if((cycles >>= 2) < RESOLUTION) {
		clockSelectBits = _BV(CS12);              // prescale by /256
		prescale_factor = 256;
	}
	else if((cycles >>= 2) < RESOLUTION) {
		clockSelectBits = _BV(CS12) | _BV(CS10);  // prescale by /1024
		prescale_factor = 1024;
	}
	else {
		// request was out of bounds, set as maximum
		cycles = RESOLUTION - 1, clockSelectBits = _BV(CS12) | _BV(CS10);
		prescale_factor = 1024;
	}
	pwmPeriod = cycles;
}

inline void TimerOne::set_period(long microseconds) {
	
	if (microseconds > 0)
		calc_period(microseconds);
	
	// ICR1 will be unchanged till set_period()
	// is called again directly or undirectly by user.
	ICR1 = pwmPeriod;       // ICR1 is TOP in p & f correct pwm mode
	
	// update clock select register
	uint8_t tmp = TCCR1B;
	tmp &= ~(_BV(CS10) | _BV(CS11) | _BV(CS12));
	TCCR1B = tmp | clockSelectBits;
}




inline void TimerOne::setPwmDuty(char pin, int duty)
{
  unsigned long dutyCycle = pwmPeriod;
  dutyCycle *= duty;
  dutyCycle >>= 10;
  if(pin == 1 || pin == 9)       OCR1A = dutyCycle;
  else if(pin == 2 || pin == 10) OCR1B = dutyCycle;
}

void TimerOne::pwm(char pin, int duty, long microseconds)  // expects duty cycle to be 10 bit (1024)
{
  if(microseconds > 0) set_period(microseconds);
  if(pin == 1 || pin == 9) {
    DDRB |= _BV(PORTB1);                                   // sets data direction register for pwm output pin
    TCCR1A |= _BV(COM1A1);                                 // activates the output pin
  }
  else if(pin == 2 || pin == 10) {
    DDRB |= _BV(PORTB2);
    TCCR1A |= _BV(COM1B1);
  }
  setPwmDuty(pin, duty);
  start();
}

void TimerOne::disablePwm(char pin)
{
  if(pin == 1 || pin == 9)       TCCR1A &= ~_BV(COM1A1);   // clear the bit that enables pwm on PB1
  else if(pin == 2 || pin == 10) TCCR1A &= ~_BV(COM1B1);   // clear the bit that enables pwm on PB2
}

inline void TimerOne::attach_interrupt(void (*isr)(), long microseconds) {
	if(microseconds > 0)
		set_period(microseconds);
	if (isr)
		isrCallback = isr;  // register the user's callback with the real ISR
	TIMSK1 |= _BV(TOIE1);   // sets the timer overflow interrupt enable bit
	sei();                  // ensures that interrupts are globally enabled
	start();
}

inline void TimerOne::detachInterrupt() {
	TIMSK1 &= ~_BV(TOIE1);  // clears the timer overflow interrupt enable bit 
}

inline void TimerOne::start() {
	TCCR1B |= clockSelectBits;
}

inline void TimerOne::stop() {
	TCCR1B &= ~(_BV(CS10) | _BV(CS11) | _BV(CS12)); // clears all clock selects bits
}

inline void TimerOne::restart() {
	TCNT1 = 0;
}

inline long TimerOne::possibly_stop_earlier(long microseconds, bool &cnt_reset, bool accurate_reply) {
	uint8_t timsk1_old = TIMSK1;
	TIMSK1 &= ~_BV(TOIE1);  // temporarily disables the timer overflow interrupt
	
	uint16_t prescaler_old = prescale_factor;
	long ms_to_run = ms_from_cnt(ICR1-TCNT1, prescale_factor);
	// TODO: Precision (Off-by-one errors)
	
	cnt_reset = false;
	
	uint16_t tcnt = TCNT1;
	
	if (prescale_factor >= 64) {
		// Should try to keep accuracy, but don't do it here.
CASSERT(F_CPU/1000000 == 16, TimerOne_cpp) // Otherwise use another value than 16
		if (ms_to_run > microseconds + 12 + prescale_factor / 16) {
			// replace it
			calc_period(microseconds);
			while (accurate_reply && tcnt == TCNT1) {}
			tcnt += accurate_reply ? 1 : 0; // We could also add 1 unconditionally
			stop();
			TCNT1 = 0;
			set_period();
			cnt_reset = true;
		}
		else {
			// In worst case the new requested stop will be ~80 us (15+64) late.
			// Don't need to do anything here.
		}
	}
	else { //prescale_factor <= 8
		// Accuracy doesn't matter
		if (ms_to_run > microseconds + 12) {
			
			calc_period(microseconds);
			if (prescale_factor > 64) {
				// This will never happen
				set_period();
				tcnt = TCNT1;
				TCNT1 = 0;
			}
			else {
				tcnt = TCNT1;
				set_period();
				TCNT1 = 0;
			}
			
			cnt_reset = true;
		}
	}
	
	long retV = ms_from_cnt(tcnt, prescaler_old);
	
	TIMSK1 = timsk1_old;   // restores the timer overflow interrupt enable bit
	// Could with small probability cause an immediate interrupt.
	return retV;
}


