/*
 *  Interrupt and PWM utilities for 16 bit Timer1 on ATmega168/328
 *  Original code by Jesse Tane for http://labs.ideo.com August 2008
 *  Modified March 2009 by Jérôme Despatis and Jesse Tane for ATmega328 support
 *  Modified June 2009 by Michael Polli and Jesse Tane to fix a bug in setPeriod() which caused the timer to stop
 *
 *  This is free software. You can redistribute it and/or modify it under
 *  the terms of Creative Commons Attribution 3.0 United States License. 
 *  To view a copy of this license, visit http://creativecommons.org/licenses/by/3.0/us/ 
 *  or send a letter to Creative Commons, 171 Second Street, Suite 300, San Francisco, California, 94105, USA.
 *
 */

#include <avr/io.h>
#include <avr/interrupt.h>

#define RESOLUTION 65536    // Timer1 is 16 bit

class TimerOne
{
  public:
  
    // properties
    uint16_t prescale_factor;
    unsigned int pwmPeriod;
    unsigned char clockSelectBits;

	// methods
	// Remark: Starting the timer will result in periodic interrupts.
	// To use the timer once you must actively call stop().
    void initialize(long microseconds=1000000);
    void start();
    void stop();
    void restart();
    void pwm(char pin, int duty, long microseconds=-1);
    void disablePwm(char pin);
    void attach_interrupt(void (*isr)(), long microseconds=-1);
    void detachInterrupt();
    void set_period(long microseconds = 0); // Doesn't affect TCNT1
	void calc_period(long microseconds);
    void setPwmDuty(char pin, int duty);
    void (*isrCallback)();
	
	/* If there was a later interrupt scheduled, it will
	 * be silently canceled in favor of the new earlier interrupt. If the
	 * new interrupt is later, the interrupts won't be changed.
	 *
	 * Input: 
	 *    microseconds: the next stop request in microseconds.
	 *    cnt_reset: here's stored wether the counter is reset or not.
	 *    accurate_replay: whether the return value shall be exact in terms
	 *         of the old replaced timeout (true), or whether the reaction
	 *         time shall be fast and (false)
	 *         or the return value shall be exact (true). [Of course the 
	 *         whole mechanics will work more accurate if the selected
	 *         periods are short. So when a 4-Second-Timeout
	 *         is running the stop-earlyer-error will be ~ 32 us, compared
	 *         to ~ 2 us when a 200-ms-timeout is running.]
	 * Return: This method returns the number of microseconds passed since the
	 * last timer start.
	 */
	long possibly_stop_earlier(long microseconds, bool &cnt_reset, bool accurate_reply = false);
	
	static long ms_from_cnt(uint16_t tcnt, uint16_t the_prescale_factor) {
		
		// Counting downward/upward doesn't matter here.
		return (long)tcnt*(long)the_prescale_factor *(F_CPU/1000000);
	}
};

extern TimerOne Timer1;
