

// © 2018 Jonas Bechtel
// Licensed under the Apache License, Version 2.0; see file <prj>/meta/LICENSE


#ifndef SIMULATOR_C
#define SIMULATOR_C


// How long a wave gets transmitted until the tranceiver is notified
// and gets active.
// If there are other events pending, the notification time span could
// be shorter than DEFAULT_SEGMENTATION_INTERVAL
//#define DEFAULT_SEGMENTATION_INTERVAL 10 MSEC
#define DEFAULT_SEGMENTATION_INTERVAL 100000 MSEC // = inactive
// (deactivated because we don't use interrupts so far.)

// If set > DEFAULT_SEGMENTATION_INTERVAL the tranceiver notification
// could be delayed if there are other events present in the next
// MAX_SEGMENTATION_INTERVAL but not in DEFAULT_SEGMENTATION_INTERVAL
#define MAX_SEGMENTATION_INTERVAL 25 MSEC

// To allow the segmentation not to intersect a waveblip
// Of course waveblip segmentation is possible anyway (if there are
// claims/events pending - they will be handled precisely).
#define LAZY_WAVEBREAK 500 USEC




#include "station.h"
#ifndef NOCODE
#define NOCODE
#include "station-names.c"
#undef NOCODE
#else
#include "station-names.c"
#endif




#ifndef NOCODE

#include <locale.h>
#include <math.h> // isnanf()

#include "sglib.h"

#include "DUMMYroute/subobject.h"
#include "AODVroute/subobject.h"

// You don't even need this three, the implicit function declaration
// would be sufficient.
#include "DUMMYapp/subobject.h"
#include "ONLYSENDapp/subobject.h"
#include "ONLYRECEIVEapp/subobject.h"
#include "FRAMESWITCHapp/subobject.h"



#include "common/trancv-user.h"

#define NOCODE
#include "read_ride.c"
#include "air.c"
#undef NOCODE

#define NOCODE
#include "multihw.c"
#undef NOCODE

#include "sim-event.h"
//#include "DUMMYapp/appobject.h"

#ifndef NOCODE
static int8_t mycomparator(uint64_t y, uint64_t x) {
	int64_t tmp = ((int64_t)y - (int64_t)x);
	int8_t retV = tmp > 0 ? 1 : tmp < 0 ? -1 : 0;
	//printf ("compare: y=%lu, x=%lu  ->  %d.\n", y, x, retV);
	return retV;
}
//#define EVENT_COMPARATOR(x, y) (y.time - x.time)
#define EVENT_COMPARATOR(x, y) mycomparator(y.time, x.time)
#define EVENT_EXCHANGER(type, events, x, y) \
	events[MAX_SIMEVENTS] = events[x]; \
	events[x] = events[y]; \
	events[y] = events[MAX_SIMEVENTS]; \
	//printf("exchanged\n")
SGLIB_DEFINE_HEAP_FUNCTIONS(simevent, event, events, i, MAX_SIMEVENTS, EVENT_COMPARATOR, EVENT_EXCHANGER)
#endif


struct simulator {
	struct station st[WBCDOlen];
	
	struct channel_in_time ch;
	
	uint64_t lowest_claim;
	int lowest_claim_station;
	
	uint64_t actual_time;
	
	// Access: wave_seq_segment[WBCDOlen * st_idx + seq_seg_nr]
	struct wave_seq_segment rxbuf[WBCDOlen * MAX_RX_SEQUENCES_PER_STATION];
	
	
	struct simevent_for_sglib queue;
};


#ifdef DEBUG_SIMEVENT
#ifndef NOCODE
static void print_queue(struct simevent_for_sglib * queue) {
	int T = 0; 
	//while (queue->events[T].time != TIME_INVALID) {
	while(T < queue->i) {
		printf ("Event %c @%'15luns\n", queue->events[T].type, queue->events[T].time);
		T++;
	}
}
#endif // -ifndef NOCODE
#endif // -ifdef SIMEVENT

// Finds the next free seq_seg for the given station
// Returns -1 if full.
static inline int wave_seq_segment(struct simulator * s, int st_idx) {
	int i;
	int offset = MAX_RX_SEQUENCES_PER_STATION * st_idx;
	for (i = 0; i < MAX_RX_SEQUENCES_PER_STATION; i++) {
		if (s->rxbuf[offset + i].waves == NULL)
			break;
	}
	if (i == MAX_RX_SEQUENCES_PER_STATION)
		return -1;
	return offset + i;
}

//#endif


#endif // -ifndef NOCODE

void clean_wave_seq_segment (struct wave_seq_segment* wss)
#ifndef NOCODE
{
	if (wss->waves == NULL) return; // Nothing to do
	
	if (!wss->waves->refcount) {
		printf("simulator.c: WARNING: this may never happen! [waves refcount zero]\n");
		return;
	}
	
	wss->waves->refcount--;
	
	if (!wss->waves->refcount) {
		//printf("freeing %p \n", wss->waves);
		free(wss->waves);
	}
	
	wss->waves = NULL;
	
	return;
}
#endif // -ifndef NOCODE
;



#define SIMULATOR(station) ((struct simulator*)((station)->dontuse1))
#define SIMULATOR_INDEX(station) ((station)->dontuse2)

uint64_t get_simtime(struct station* station)
#ifndef NOCODE
{
	//fprintf(stderr, "station is @ %p.\n", station);
	//struct simulator* sim = SIMULATOR(station);
	//fprintf(stderr, "dontuse1 = %p.\n", sim);
	//fprintf(stderr, "sim = %p.\n", sim);
	return SIMULATOR(station)->actual_time;
}
#endif // -ifndef NOCODE
;


time_int next_action(struct simulator * s)
#ifndef NOCODE
{
	time_int next_time = s->lowest_claim;
	if (!sglib_simevent_is_empty(&s->queue)) {
		 // peek first element
		struct event e = sglib_simevent_first_element(&s->queue);
		if (e.time < next_time)
			next_time = e.time;
	}
	
	if (next_time == CLAIM_INVALID) {
		next_time = TIME_MAX;
	}
	return next_time;
}
#endif // -ifndef NOCODE
;


/*inline void do_place_claim(
			struct simulator * s, struct station* st, uint64_t claimtime) {
	st->claimvar = claimtime;
	
	if (claimtime < st->lowest_claim) {
		st->lowest_claim = claimtime;
		st->lowest_claim_station = SIMULATOR_INDEX(st);
	}
}*/

void claim(struct station * st, uint64_t claimtime)
#ifndef NOCODE
{
#ifdef DEBUG_SIMEVENT
	printf("Simulator: got claim for %c@%'ld.\n", st->name, claimtime);
#endif
	
	if (st->claimvar <= claimtime)
		return;
	
	st->claimvar = claimtime;
	
	if (claimtime < SIMULATOR(st)->lowest_claim) {
		SIMULATOR(st)->lowest_claim = claimtime;
		SIMULATOR(st)->lowest_claim_station = SIMULATOR_INDEX(st);
	}
}
#endif // -ifndef NOCODE
;


void quit(struct station * st, const char* reason)
#ifndef NOCODE
{
	printf("Simulator: got \"normal\" quit request for %c (because %s)\n", st->name, reason);
	
	struct event ev;
	ev.time = 0;
	ev.type = '0';
	ev.station = st->name;
	sglib_simevent_add(&SIMULATOR(st)->queue, ev);
}
#endif // -ifndef NOCODE
;

void error(struct station * st, const char* reason)
#ifndef NOCODE
{
	printf("Simulator: got ERROR quit request for %c (because %s)\n", st->name, reason);
	
	struct event ev;
	ev.time = 0;
	ev.type = 'E';
	ev.station = st->name;
	sglib_simevent_add(&SIMULATOR(st)->queue, ev);
}
#endif
;



#ifndef NOCODE

static void feed_wave_segment_to_trancv(struct simulator * s, time_int starttime) {
	
	time_int breaktime = next_action(s);
	time_int lazy_wavebreak = 0;
//	if (breaktime > MAX_SEGMENTATION_INTERVAL + get_simtime(s)) {
//		breaktime = DEFAULT_SEGMENTATION_INTERVAL + get_simtime(s);
//		lazy_wavebreak = LAZY_WAVEBREAK;
//	}
	
	if (breaktime == starttime) {
		// Unfortunately it didn't work to use events to trigger an aircombine
		// call so we have to put an event which is special: it will be called
		// as last event of actual simtime.
		
		struct event e;
		e.time = starttime;
		e.type = 'R';
		e.station = 0; // All stations
		
		sglib_simevent_add(&s->queue, e);
		//printf("Added R event. lowest_claim = %'lu, e.time = %'lu.\n", s->lowest_claim, e.time);
		
		return;
	}
	
	int i;
	
	// Now put it into the storage space for every receiving station
	for (i = 0; i < WBCDOlen; i++) {
		/*if (i == st_idx)
			// TODO: "continue" is wrong in case of far reflections.
			continue;*/
		// NOTE: We let out continue because some waveblips wouldn't
		// be processed and would lead to old timestamps which puzzle
		// the program.
		// TODO: maybe fake-process theses uneeded waveblips for the
		//       sake of speed?
		
		//printf("n2:%d\n", MAX_RX_SEQUENCES_PER_STATION * i);
		
		
		
		struct wave_seq_segment wss2 = 
		aircombine_new(&(s->rxbuf[MAX_RX_SEQUENCES_PER_STATION * i]),
					MAX_RX_SEQUENCES_PER_STATION,
					starttime, // It would be a bug if there was anything
								// before starttime present in rxbuf, but
								// we don't check for this bug.
					breaktime, lazy_wavebreak);
		// TODO: add noise!
		
		// TODO: It would be better to request a minimum RSSI level and
		//       notify the receiver once the RSSI level supasses the 
		//       given minimum.
		
#ifdef DEBUG_SIMEVENT_DETAIL
		// TODO: make print_wave_seq_segment() out of this:
		printf("wss2.starttime=       %'15luns\n", wss2.starttime);
		printf("wss2.segment_duration=%'15uns\n", wss2.segment_duration);
		printf("wss2.num_second_element=%'13u\n", wss2.num_second_element);
		printf("wss2.waves->capacity=   %'13u\n", wss2.waves->capacity);
		printf("wss2.first_element.duration=%'13u      .amplitude=%f\n", wss2.first_element.duration, wss2.first_element.amplitude);
		int T = 0;
		while (T < wss2.waves->capacity) {
			printf("wss2.waves->seq[%d].duration=%'13u      .amplitude=%f\n", T, wss2.waves->seq[T].duration, wss2.waves->seq[T].amplitude);
			if (isnanf(wss2.waves->seq[T].amplitude)) break;
			T++;
			// Don't count duration as it's simply a display stuff.
		}
#endif
		
		
		if (wss2.waves == NULL) {
			// No news. (So probably a reflected wave which not reaches
			// every station)
			continue;
		}
		
		
		struct wave_seq_segment * wssp = malloc(sizeof(struct wave_seq_segment));
		if (wssp == NULL) {
			printf ("simulator.c: ERROR: couldn't allocate memory! (send handling).\n");
			continue;
		}
		*wssp = wss2;
		
		
		// In order the receiver not to fake with time we notify it at the end
		// of transmission.
		
		struct event ev;
		ev.type = 'R';
		ev.time = wss2.starttime + wss2.segment_duration;
		ev.station = WBCDO[i];
		ev.wssp = wssp;
		
		sglib_simevent_add(&s->queue, ev);
		
	} // for (i = 0; i < WBCDOlen; i++)
	
}
#endif // -ifndef NOCODE


void emit_wave_segment_to_air(struct station * st,
		struct wave_sequence * waves, time_int delay
)
#ifndef NOCODE
{
	if (!waves)
		return;
	
	struct simulator * s = SIMULATOR(st);
	
	waves->source = st->name;
	
#ifdef DEBUG_AIR
	printf("Got some data from %c for %lu ms (%'lu ns).\n", st->name, 
		   waves->duration / 1000000, waves->duration);
#endif
	//printf("waves->skip = %d,   waves->capacity = %d\n", waves->skip, waves->capacity);
	assert2 (waves->skip <= waves->capacity, "Something about the waves capacity is wrong!\n");
	int i;
	for (i = waves->skip; i < waves->capacity; i++) {
		if (__isnanf(waves->seq[i].amplitude))
			break;
		
		//printf("waveblip: %6d ns @%9d Hz with %f mW and phase %d \n",
		//		waves->seq[i].duration, waves->seq[i].frequency,
		//		waves->seq[i].amplitude, waves->seq[i].phase);
	}
	// TODO: maybe check duration? (But we cann't check everything everytime)
	
	
	
	
	/*
	struct event ev;
	ev.time = s->actual_time + answer.do_delay + waves->segmentation;
	ev.type = '%';
	sglib_simevent_add(&(s->queue), ev);*/
	
	
	uint64_t starttime = s->actual_time + delay;
	
	if (s->ch.total_time_s <= starttime/1000000000) {
		fprintf(stderr, "Too late.\n");
		return;
	}
	
	/*TODO: 1 SEC is wrong access */
	uint32_t line_from  = s->ch.mapper[starttime / (1 SEC)].from;
	uint32_t line_break = s->ch.mapper[starttime / (1 SEC)].to + 1;
#ifdef DEBUG_AIR
	// The more near those two values are, the more efficient is the search then
	printf("line_from = %d, line_break = %d\n", line_from, line_break);
#endif
	int j;
	for (j = line_from; j < line_break; j++) {
		// In the for loop we are looking for the channel lines which 
		// are related to the recent time.
		
		// But there may be (time-)unrelated lines present
		if (starttime < s->ch.chtimespecs[j].time 
			|| s->ch.chtimespecs[j].time + s->ch.chtimespecs[j].duration MSEC < starttime )
			// This is such an unrelated line.
			continue;
		
		
		// Now put it into the storage space for every receiving station
		for (i = 0; i < WBCDOlen; i++) {
			/*if (i == st_idx)
				// TODO: "continue" is wrong in case of far reflections.
				continue;*/
			// NOTE: We let out continue because some waveblips wouldn't
			// be processed and would lead to old timestamps which puzzle
			// the program.
			// TODO: maybe fake-process theses uneeded waveblips for the
			//       sake of speed?
			
			
			int chsel;
			if (i <= SIMULATOR_INDEX(st))
				// i == st_idx is irrelevant as the result is the same.
				chsel = get_chsel(i, SIMULATOR_INDEX(st));
			else
				chsel = get_chsel(SIMULATOR_INDEX(st), i);
			
			
			if (s->ch.chtimespecs[j].ch[chsel].dist == 0xFF) {
				continue; // Thanks, valgrind!
			}
			
			
			
			int n = wave_seq_segment(s, i);
			if (n == -1) {
				// rescue try: combine the slices which are present.
				printf("simulator.c: WARNING: the storage for the"
						" to-be-combined is too small.\n");
				
				bool still_error = false;
				struct wave_seq_segment wss = aircombine_new(
							&(s->rxbuf[MAX_RX_SEQUENCES_PER_STATION * i]),
							MAX_RX_SEQUENCES_PER_STATION,
							starttime, // s->actual_time, // 0, [Irrelevant
							// what stands here. Actually there should be no
							// wave at all before s->actual_time, otherwise
							// there's a big bug in this proram]
							TIME_MAX, 0);
				
	printf("waves_duration from %c = %lld ms (%'llu ns).\n", st->name,
		   waves->duration / 1000000, waves->duration);
				
				n = wave_seq_segment(s, i);
				if (n == -1) {
					// Some calculation made for nothing.
					clean_wave_seq_segment(&wss); // freeing wss->waves
					
					still_error = true;
				}
				else {
					// There has been freed one slice.
					s->rxbuf[n] = wss;
					
					// Allocate slice for the actual task.
					int n2 = wave_seq_segment(s, i);
					
					// Try again to find a place for the new slice:
					if (n2 == -1) {
						// Some calculation made for nothing.
						// Freeing the waves and the rxbuf place.
						clean_wave_seq_segment(&s->rxbuf[n]);
						still_error = true;
					}
					n = n2;
				}
				
				if (still_error) {
					
#define ERRSLICESTR "Simulator ERROR [deferred]: No more slices for waveseq_segments free"
					printf(ERRSLICESTR " @station %c!\n"
					"Consider raising MAX_RX_SEQUENCES_PER_STATION.", WBCDO[i]);
					error(&s->st[i], ERRSLICESTR "!\n");
					continue;
				}
			}
			//printf("XX:n=%d\n", n);
			
			struct wave_seq_segment wss;
			
			wss.starttime = starttime;
			//printf ("Setting starttime %llu.\n", starttime);
			wss.first_element = waves->seq[waves->skip];
			wss.num_second_element = waves->skip + 1;
			
			
			// Let's look how much time is available
			// (Include connection time from database)
			
			wss.segment_duration = waves->duration;
			
			waves->refcount++;
			wss.waves = waves;
			
			
			// Notify the receiver's routines.
			// This is the point where multi-threading could start.
			
			// For the timing look at the read-in channel lines.
			
			// implant the distance and damp
			wss.airinfo.dist = 0;
#ifdef DEBUG_AIR_DETAIL
			printf("j=%d,chsel=%d, dist = %d   dampx = %f\n", j, chsel, 
				   s->ch.chtimespecs[j].ch[chsel].dist,
				   s->ch.chtimespecs[j].ch[chsel].damp);
#endif
			wss.airinfo.damp = s->ch.chtimespecs[j].ch[chsel].damp;
			wss.starttime += s->ch.chtimespecs[j].ch[chsel].dist * 100;
			
			//printf("s->rxbuf[%d].waves = %p.\n", n, s->rxbuf[n].waves);
			s->rxbuf[n] = wss;
			
			
			
		}
	}
	
	
	
	// After transmission, station may want to do something
	// Using event mechanism for this task.
	struct event finished;
	finished.time = starttime + waves->duration;
	finished.type = '$';
	finished.station = st->name;
	sglib_simevent_add(&s->queue, finished);
	
	// handle the waves to the receivers
	
	feed_wave_segment_to_trancv(s, starttime);
	
	// Remove the wave sequence when no needed. Construct wss for that.
	struct wave_seq_segment wss = {.waves = waves};
	clean_wave_seq_segment(&wss);
	
}
#endif // -ifndef NOCODE
;

#ifndef NOCODE


// This never happens??
static void handle_answer(struct simulator * s, int st_idx, struct trigger_corroboration answer) {
#ifdef DEBUG_SIMEVENT
	printf ("SH000\n");
#endif
	
	switch(answer.type) {
	case to_air:
		// TODO: All this code under to_air could be moved to an own (inline)
		// function for readability.
		if (!answer.to_air) {
			printf("Simulator: ERROR: I got zero wave answer from station %c.\n", WBCDO[st_idx]);
			quit(NULL, "Simulator: I got zero wave answer from some station.");
			break;
		}
		emit_wave_segment_to_air(&s->st[st_idx], answer.to_air, answer.do_delay);
		
		break;
	case Quit:
		quit(&s->st[st_idx], answer.reason);
		break;
	case Error:
		error(&s->st[st_idx], answer.reason);
		break;
	case nop:
		break;
	default:
		printf("Simulator: unknown answer type after claimed trigger to station %c!\n",
				WBCDO[st_idx] /*s.st[st_idx].name*/);
	}
#ifdef DEBUG_SIMEVENT
	printf ("SH999\n");
#endif
}


// To be invoked by the stations
static void place_extra_answer(struct station * st, struct trigger_corroboration tc) {
	handle_answer(SIMULATOR(st), SIMULATOR_INDEX(st), tc);
}


int main() {
	int res = 0;
	
	struct simulator s;
	s.lowest_claim = CLAIM_INVALID;
	s.lowest_claim_station = 0;
	
	
	struct simevent_for_sglib* queue = &s.queue;
	sglib_simevent_init(queue);
	//printf("max=%d, i=%d\n", MAX_SIMEVENTS, queue.i); // max=100, i=0
	
	// Set time to zero
	s.actual_time = 0;
	
	// Thousands' commas and decimal points in any environment.
	setlocale(LC_NUMERIC, "en_US");
	// Remark: you _have_ to call setlocale(LC_NUMERIC, <anything>) in order
	// to get thousands' seperators at all!
	
	
	int i; // will be reused
	
	// Invalidate all temporary wave memory.
	for (i = 0; i < (sizeof(s.rxbuf)/sizeof(s.rxbuf[0])); i++) {
		s.rxbuf[i].waves = NULL;
	}
	
	
	// Initialize all stations
	for (i = 0; i < WBCDOlen; i++) {
		s.st[i].dontuse1 = &s;
		s.st[i].dontuse2 = i;
		s.st[i].name = WBCDO[i]; // Select 'W', 'B', 'C', 'D' or 'O' from the station list
		
		
		// Hardware abstraction layer (when working: ulog, timer, ...)
		s.st[i].mgs = init_multihw(&s.st[i]);
		if (s.st[i].mgs == NULL) {
			printf("Error initializing the HAL (multihw) of station %c.\n",
				WBCDO[i]);
		}
		
		
		//TODO: Here will happen a route claim. -> event?
		if (res != OBJECT_START_OK) {
			printf("Error initializing the routing instance of station %c.\n",
				WBCDO[i]);
			goto fail;
		}
		
		// Application
		/*if (s.st[i].name == 'W' || s.st[i].name == 'B' || s.st[i].name == 'D') {
			//s.st[i].app = new_DUMMYapp(&s.st[i]);
			s.st[i].app = new_ONLYSENDapp(&s.st[i]);
			if (!s.st[i].app) goto fail;
			
			// This will also initialize the tranceiver.
			res = start_ONLYSENDapp(s.st[i].app);
			//res = start_DUMMYapp(s.st[i].app);
		}
		else {
			s.st[i].app = new_ONLYRECEIVEapp(&s.st[i]);
			if (!s.st[i].app) goto fail;
			
			// This will initialize the tranceiver.
			res = start_ONLYRECEIVEapp(s.st[i].app); 
		}*/
		
		s.st[i].app = new_FRAMESWITCHapp(&s.st[i]);
		if (!s.st[i].app) goto fail;
		// This will also initialize the tranceiver and the routing.
		res = start_FRAMESWITCHapp(s.st[i].app);
		
		if (res != OBJECT_START_OK) {
			printf("Error initializing the app of station %c.\n",
				WBCDO[i]);
			goto fail;
		}
		
		
		s.st[i].claimvar = CLAIM_INVALID;
		s.st[i].claim = claim;
		s.st[i].quit = quit;
		s.st[i].error = error;
		s.st[i].place_extra_answer = place_extra_answer;
	}
	printf("Simulator main: pre-initialized %d stations.\n", i);
	
	struct event ev;
	ev.time = 0;
	ev.type = '1';
	ev.station = 0;
	
	sglib_simevent_add(queue, ev);
	
	
	
	s.ch = read_ride("../output.Gen._ride");
	
	if (s.ch.chtimespecs)
		printf("Simulator main: Read in %d rows of channel data "
				"(total %d seconds).\n",
				s.ch.speclines, s.ch.total_time_s);
	else {
		printf("Simulator main: failed to read data.\n");
		goto fail;
	}
	
	map_channeltimes(&s.ch);
	if (!s.ch.mapper) {
		printf("Simulator main: failed to map channel data to time array.\n");
		goto fail;
	}
	
	//for (i = 0; i < s.ch.total_time_s; i++)
	//	printf("    @%4ds: Lines %4d->%4d.\n", i, s.ch.mapper[i].from, s.ch.mapper[i].to);
	
	
	
	// Setting up app (~ Benchmark)
	
	
	
	s.actual_time = 0;
	uint8_t reason = 1; // Loop quit reason: 0 = quit, 2 = Error, (1 = running)
	
	// EVENT LOOP
	while (true) {
		// There are two types of "events":
		// 1. the actual events
		// 2. the claims (= triggers).
		// We check first for the basic events (Quit, Error).
		// Then we decide whether to proceed with claims or with errors.
		//     Decision will be taken on the earlier of both.
		
		
		// Claims may be altered by caller, events are unchangeble once
		// put into the loop.
		
		
		
		if (sglib_simevent_is_empty(queue)) {
			
			if (s.ch.total_time_s <= s.lowest_claim/1000000000) {
				reason = 'T';
				break;
			}
			
			goto process_claim;
		}
		
		struct event e = sglib_simevent_first_element(queue);
		if (e.type == 'R' && e.station == 0 && queue->i > 1) {
			
			if (queue->events[1].time == e.time) {
				// Postpone 'R' event of equal time.
				queue->events[0] = queue->events[1];
				queue->events[1] = e;
				e = queue->events[0];
			
				if (e.type =='R' && e.station == 0) {
					//printf("delete\n");
					// equal element, so we can remove one of both first ones.
					sglib_simevent_delete_first(queue);
				}
			}
		}
		
		
		if (e.type == '0') {
			reason = 0;
			printf("Simulator: carrying out quit request of station %c.\n", e.station);
			break;
		}
		if (e.type == 'E') {
			reason = 2;
			printf("Simulator: carrying out error request of station %c.\n", e.station);
			break;
		}
		
#ifdef DEBUG_SIMEVENT
		print_queue(queue);
#endif
		
		if (s.ch.total_time_s <= e.time/1000000000
				&& s.ch.total_time_s <= s.lowest_claim/1000000000) {
			reason = 'T';
			break;
		}
		
		
		if (e.time == s.lowest_claim && e.type == 'R' && e.station == 0)
			goto process_claim;
		if (e.time <= s.lowest_claim)
			goto process_event;
		
process_claim:
		if (s.lowest_claim == CLAIM_INVALID)
			// Leave loop. This will exit the program.
			break;
		if (s.st[s.lowest_claim_station].claimvar < s.actual_time) {
#ifdef DEBUG_SIMEVENT
			printf("Simulator main: WARNING: processing a claim which is in past!\n"
					"   (actual time: %ld, claim time: %'ld)\n", s.actual_time,
					s.st[s.lowest_claim_station].claimvar);
#endif
		}
		else {
			s.actual_time = s.st[s.lowest_claim_station].claimvar;
#ifdef DEBUG_SIMEVENT
			printf("Info: proceeding to the time %'ld. (claim by station %c)\n",
				s.actual_time, WBCDO[s.lowest_claim_station]);
#endif
		}
		
		//printf("MCCCC\n");
		int idx_tmp = s.lowest_claim_station;
		assert(WBCDO[idx_tmp] == s.st[idx_tmp].name); // Safety check
		
		// Now the lowest claim has been triggered, put the next claim in effect.
		s.st[s.lowest_claim_station].claimvar = CLAIM_INVALID;
		s.lowest_claim = s.st[0].claimvar;
		s.lowest_claim_station = 0;
		for (i = 1; i < WBCDOlen; i++) {
 			if (s.st[i].claimvar < s.lowest_claim) {
				s.lowest_claim = s.st[i].claimvar;
				s.lowest_claim_station = i;
			}
		}
		
		//event ev; // TODO: is this carried out? Don't we need this?
		ev.time = 0xFEDCBA9876543210; // s.actual_time;
		ev.type = 'C';
		ev.station = 0; // Doesn't matter. Station knows it's own name.
		//printf("beep. %d %d\n", idx_tmp, s.lowest_claim_station);
		struct trigger_corroboration answer;
		answer = s.st[idx_tmp].app->notify(s.st[idx_tmp].app, ev);
#ifdef DEBUG_SIMEVENT
		printf("MDDDD\n");
#endif
		if (answer.type != nop)
			assert(0);
			// TODO: if answer is something relevant, it has to be respected!
		
		goto loop_end;
		
process_event:
		// TODO: first_element geht mit queue.events[0].
		// delete_first ist nicht so super implementiert.
		// Man könnte stattdessen eine eigene Methode implementieren,
		// die die ersten N Einträge löscht (einfach alles um N nach oben
		// schieben und i um N reduzieren.)
		
		sglib_simevent_delete_first(queue);
		
#ifdef DEBUG_SIMEVENT
		printf("Got event %c@%'lu from queue.\n", e.type, e.time);
#endif
		s.actual_time = e.time;
		
		time_int next_time = s.lowest_claim;
		if (!sglib_simevent_is_empty(queue)) {
			struct event etmp = sglib_simevent_first_element(queue);
			if (etmp.time < next_time)
				next_time = etmp.time;
		}
		
		
		//printf("AAAAA\n");
		if (e.type == '1') {
			for (i = 0; i < WBCDOlen; i++) {
				//printf (" invoking %p = app->system_startup()\n", s.st[i].app->system_startup);
				s.st[i].app->system_startup(s.st[i].app, s.actual_time);
			}
		}
		else if (e.type == '$') {
			i = get_WBCDO_idx(e.station);
			if (s.st[i].trancv->data_sent)
			s.st[i].trancv->data_sent(s.st[i].trancv);
		}
		else if (e.type == 'R' && e.station == 0) {
			feed_wave_segment_to_trancv(&s, e.time);
		}
		else if (e.type == 'R') {
			
			
			i = get_WBCDO_idx(e.station);
			
			// At first give the wave to the station
#ifdef DEBUG_SEND_RECEIVE
			printf("Simulator(%c): Giving %'d ns of wave to receiver.\n",
						s.st[i].name, e.wssp->segment_duration);
#endif
			/*printf(stderr, "wssp->waves->refcount = %u\n", e.wssp->waves->refcount);
			fflush(stdout);*/
			
			// Expecting tranceiver to clean_wave_seq_segment().
			s.st[i].trancv->next_wave_segment(s.st[i].trancv, *e.wssp);
			// If it doesn't it is responsible for a big memory leak.
			
			/*printf(stderr, "wssp = %p\n", e.wssp);
			fflush(stdout);
			if (e.wssp)
				printf(stderr, "wssp->waves = %p\n", e.wssp->waves);
			if (e.wssp && e.wssp->waves)
				printf(stderr, "wssp->waves->refcount = %u\n", e.wssp->waves->refcount);*/
			
			/*if (!e.wssp->waves->refcount)
				free(e.wssp);
			else {
				printf("simulator(%c): WARNING: tranceiver didn't use full wave data.\n", s.st[i].name);
				clean_wave_seq_segment(e.wssp);
			}*/
			
			free(e.wssp);
			
			// Check whether there's further wave data.
			bool found = false;
			int num = MAX_RX_SEQUENCES_PER_STATION * i;
			int j;
			for (j = 0; j < MAX_RX_SEQUENCES_PER_STATION; j++) {
				if (!s.rxbuf[num+j].waves) continue;
				
				//printf("rxbuf[%d].starttime=%lu, rxbuf[%d].airinfo.dist=%u*100ns.\n", i, rxbuf[i].starttime, i, rxbuf[i].airinfo.dist);
				if (s.rxbuf[num+j].starttime == e.time) {
					found = true;
				}
			}
			// TODO
			//XXX
			if (found) {
				feed_wave_segment_to_trancv(&s, e.time);
			}
		}
		
		//printf("BBBBB\n");
		
loop_end:
		true;
	}
	
	char* strreason;
	if (reason == 0)
		strreason = "Quit requested by station";
	else if (reason == 2)
		strreason = "Quit requested by station with ERROR status";
	else if (reason == 'T')
		strreason = "No more driving on input map = air data has ended";
	else if (sglib_simevent_is_empty(queue))
		strreason = "Queue empty";
	else
		strreason = "Unknown reason";
	printf("Event loop quit (%s).\n", strreason);
	
	res = 0;
end:
	for (i = 0; i < WBCDOlen; i++) {
		// App will/should destroy routing instance.
		// Routing instance then should display collected dBms
		s.st[i].app->destroy(s.st[i].app);
	}
	return res;
fail:
	res = 1;
	goto end;
}

#endif // ifndef NOCODE



#endif // ifndef SIMULATOR_C

