

% © 2018 Jonas Bechtel
% Licensed under the Apache License, Version 2.0; see file <prj>/meta/LICENSE



% This is to be sourced by Genride.m



% Explanation for station and other abbreviations in /Hardware/STATIONSNAMEN.txt
% Front<- W B C D O <-End
% Policies: K P (defined below in DYNAMICS)
% Surroundings: U Y (used in XXX TODO)





% DYNAMICS

% Assuming following dynamics, simplified for programming's sake:
% (Not taken into consideration: lower front-end-distance on lower speeds,
%  overall convoy speed change, breaks for drinking/toilet, traffic lights)
% 
% W and O have always a distance of 900 m. W demarks the beginning of the convoy
%    and O the end of the convoy. The convoy is going with 3.5 m/s.
% (Point of reference for station positions is the position of W)
% B is always changing between 275 m and 325 m (begins with 275 m and falls
%    back to 325 m in 150 seconds. Then goes to 275 m in 100 seconds.)
% C is "corking" an then going to front again (= 30 m). This cycle begins with
%    a wait
%    of 150 meters, then corking (v=0) starts. When reached the end (45 m in
%    front of O, thats pos(O) PLUS 45 m), C proceeds
%    to the front again in 412.5 seconds (v = 2 m/s + 3.5 m/s for 825 m)
%    On countryside C is always 40 m behind W, but after 2000 m wait it will
%    begin to cork.
% For D, there exist two policies K and P. (-> /BUCHSTABENKUERZEL.txt)
% D is also "corking" in policy K (corK). In urban region, D corks 38 seconds
%    after C corks. After reaching the end, D proceeds to the front (= 30 m) in
%    580 seconds (v = 1.5 m/s + 3.5 m/s). The cycle starts again. If C is not
%    present when D arrives in front, D will part after 19 seconds. On
%    countryside D remains between 600 m and 650 m, cycling in the same way B
%    does.
% In policy P (static Place), D remains between 600 m and 650 m, cycling in
%    the same way B does.

% Remark: the positions of the bicycles are not meant to be accurate as
% described above but this program simply wants to create a half-decent
% base for simulations.

% Remark: despite of the fact that in real world the positions continuously 
%	change and that the ride format can output sub-second accuracy timings
%	the main assumption for the dynamics is that every cyclists stands still
%	for one second and then jumps to his/her next position in an instant.


% gen_dynamics*()
% 
% stoptime is the time for which no more row is created (in other words stoptime
% is the desired number of rows)
% countrYtime is the time when the convoy is in the rural area (in other word
%  when the convoy is not in the Urban area. Every row of the countrYtime
%  matrix denotes the beginning time (inclusive) and the end time (exclusive).
% 
% returns: an array of positions. The width is 5 where the leftmost entry (or,
% more precisely, the entry with the lowest index) stands for station/cyclist W
% and the rightmodst entry stands for station/cyclist O.
% Every row of the returned array stands for one second (so for the beginning
% of the second, just when the cyclist will rest for one second before he/she
% "jumps" to his/her next position)


function x = gen_dynamics_K(stoptime, countrYplace)
	x = gen_dynamics("K", stoptime, countrYplace);
	return;
endfunction

function x = gen_dynamics_P(stoptime, countrYplace)
	x = gen_dynamics("P", stoptime, countrYplace);
	return;
endfunction

% policy is "K" or "P".
function x = gen_dynamics (policy, stoptime, countrYplace)
	% W and O are independent on policy and B/C/D.
	
	% As far as we don't need parametrizable dynamics, all numbers are
	% magic numbers!
	
	x = zeros(0,5);
	
	if (sum(typeinfo(stoptime) != "scalar"))
		fprintf("Genride:gen_dynamics*(): ERROR: stoptime is not a number!\n");
		return
	endif
	if (sum(typeinfo(countrYplace) != "matrix"))
		fprintf(["Genride:gen_dynamics*(): ERROR: countrYplace is not a "...
			"matrix!"]);
		return
	endif
	if (size (countrYplace)(2) != 2)
		fprintf(["Genride:gen_dynamics*(): ERROR: countrYplace matrix shall "...
			"be like this:\n    [start_cntr1, end_cntr1;"...
			"\n     start_cntr2, end_cntr2;"...
			"\n     ...]\n"]);
		return
	endif
	
	speed_W = 3.5; % m/s
	
	for i = 1 : stoptime
		% This could be more efficient but won't be called too frequently.
		posW = speed_W*(i-1);
		posO = max(0, posW - 900);
		x(i,:) = [posW, -1, -1, -1, posO]; % W, O
	endfor
	
	
	
	% B is only dependent on W.
	% (!) Reused in D, Policy K
	% (!) Reused in D, Policy P
	Boffset = 275; % number of meters which B cycle happens behind W
	
	time_Bstart = find(x(:,1) > Boffset, 1) - 1; % ~= num_zeros
	x(1:time_Bstart,2) = 0;
	remaintime = size(x)(1) - time_Bstart;
	
	time_to_high = 150; % seconds    % (high meint _hintere_ Position)
	distance_to_high = 50; % meters
	time_to_low = 100; % seconds
	Bfunction1 = [0:time_to_high - 1]' .* distance_to_high/time_to_high;
	Bfunction2 = [time_to_low : -1 : 1]' .* distance_to_high/time_to_low;
	Bfunction_combined = [Bfunction1;Bfunction2];
	
	% Ansatz: wir hängen den Durchlauf nochmal an den Durchlauf an (und
	% verdoppeln dadurch die Größe des Durchlaufs. Das passiert in sehr
	% wenigen Iterationen bis hin zu sehr großen Durchlaufgrößen. 
	% Die durchlaufgröße wächst exponentiell.
	times = ceil (log2(remaintime/size(Bfunction_combined)(1)));
	for i = 1 : times
		Bfunction_combined = [Bfunction_combined;Bfunction_combined];
	endfor
	
	% Bfunction was set up relative to the dynamic position of W.
	x(time_Bstart+1:time_Bstart+remaintime,2) = ...
			   x(time_Bstart+1:time_Bstart+remaintime,1)  ...
			.- Bfunction_combined(1:remaintime,:) ...
			.- Boffset; % B
	
	
	
	% C is dependent on W, O and on Y/U time
	% TODO: scene Y and Y/U mixture.
	% (!) Reused in D, Policy K
	
	time_Cstart = find(x(:,1) > 30, 1) - 1; % ~= num_zeros
	x(1:time_Cstart,3) = 0;
	
	newtime = time_Cstart + 1;
	lastpos = 0;
	
	
	while newtime <= stoptime
		% INFO: not stable when the end of the convoy is nearer than 150 m
		%	 behind W (and other things may happen as well.)
		
		% Now C is in the front and waits there for 150 m.
		% (soz. mit der Spitze mitschwimmen)
		
		time_to_cork = round(150 / speed_W);
		time_to_cork = min(stoptime+1 - newtime, time_to_cork);% Access safety.
		
		nc = next_countryside(countrYplace, lastpos);
		if (nc && nc(1,2) < speed_W*time_to_cork + lastpos)
			% Recalculate this with longer scope (as there are not so much
			% crossings in landscape.)
			
			time_to_cork = round(min(2000, nc(1,3)-nc(1,2)) / speed_W);
			time_to_cork = min(stoptime+1-newtime,time_to_cork); %Access safety.
		endif
		
		Cfunction = [1 : time_to_cork]' .* speed_W;
		x(newtime:newtime+time_to_cork - 1, 3) = Cfunction .+ lastpos;
		
		lastpos = lastpos + Cfunction(time_to_cork,1);
		newtime = newtime + time_to_cork;% Neuer Index für den Zugriff auf x.
		if (newtime > stoptime)
			break
		endif
		
		% Now we have to find out when the end of the convoy reaches the
		% position where C is standing.
		time_to_cnend = find(x(newtime:stoptime,5) + 45 > lastpos, 1);
		if (size(time_to_cnend)(1) == 0)
			time_to_cnend = stoptime+1 - newtime;% Access safety.
		endif
		
		
		Cfunction = zeros(time_to_cnend, 1);
		% C reached the "end" (45 m before) by standing still.
		x(newtime:newtime+time_to_cnend-1,3) = Cfunction .+ lastpos;
		lastpos = lastpos + Cfunction(time_to_cnend,1);
		newtime = newtime + time_to_cnend;
		if (newtime > stoptime)
			break
		endif
		
		
		% Now C proceeds to the front (30 m after W) again.
		
		% Function of W: posW-30m + 3.5 m/s * t  (point 30 m after W)
		% Function of C: posC     + 5.5 m/s * t
		% Intersection: 
		%   posW-30m + 3.5 m/s * t = posC     + 5.5 m/s * t  | - 3.5 m/s * t
		%   posW - 30 m            = posC     + 2   m/s * t  | - posC |  / 2m/s
		%  (posW - 30 m - posC) / (2m/s)  =               t   
		time_to_front = round(1/2.0 * (x(newtime-1,1)-30-lastpos) - 1);
		time_to_front = min(stoptime+1 - newtime, time_to_front);% Access safety
		
		fastspeed = 5.5; % m/s
		Cfunction = [1 : time_to_front ]' .* fastspeed;
		x(newtime:newtime+time_to_front-1,3) = Cfunction .+ lastpos;
		lastpos = lastpos + Cfunction(time_to_front,1);
		newtime = newtime + time_to_front;
		
		% Now C has proceeded to the front. Here the loop begins again.
		
		
	endwhile
	
	
	
	if (policy == "P")
		% D is only dependent on W. (Policy P)
		% (!) Code duplication from B
		
		Doffset = 600; % number of meters which D cycle happens behind W
		
		time_Dstart = find(x(:,1) > Doffset, 1) - 1; % ~= num_zeros
		x(1:time_Dstart,4) = 0;
		remaintime = size(x)(1) - time_Dstart;
		
		time_to_high = 150; % seconds
		distance_to_high = 50; % meters
		time_to_low = 100; % seconds
		Dfunction1 = [0:time_to_high - 1]' .* distance_to_high/time_to_high;
		Dfunction2 = [time_to_low : -1 : 1]' .* distance_to_high/time_to_low;
		Dfunction_combined = [Dfunction1;Dfunction2];
		
		% Ansatz: wir hängen den Durchlauf nochmal an den Durchlauf an (und
		% verdoppeln dadurch die Größe des Durchlaufs. Das passiert in sehr
		% wenigen Iterationen bis hin zu sehr großen Durchlaufgrößen. 
		% Die durchlaufgröße wächst exponentiell.
		times = ceil (log2(size(x)(1)/size(Dfunction_combined)(1)));
		for i = 1 : times
			Dfunction_combined = [Dfunction_combined;Dfunction_combined];
		endfor
		
		% Dfunction was set up relative to the dynamic position of W.
			x(time_Dstart+1:time_Dstart+remaintime,4) = ...
			   x(time_Dstart+1:time_Dstart+remaintime,1)  ...
			.- Dfunction_combined(1:remaintime,:) ...
			.- Doffset; % D
		
	else
		% D is dependent on W, C, O and on Y/U time (Policy K) (TODO)
		% (!) Code duplication from C (to some extend)
		
		% TODO: scene Y and Y/U mixture.
		
		time_Dstart = find(x(:,1) > 30, 1) - 1; % ~= num_zeros
		x(1:time_Dstart,4) = 0;
		
		newtime = time_Dstart + 1;
		lastpos = 0;
		
		state = "static";
		% (static, seekpos, corking, tofront, pendule)
		
		
		while newtime <= stoptime
			% INFO: not stable when the end of the convoy is nearer than 150 m
			%	 behind W (and other things may happen as well.)
			% INFO: not stable when rural <-> urban changes frequently.
			
			if (strcmp(state, "static"))
				% Now D is in the front and waits for C corking and for
				% additionl 38 s. (or 19 s in total if C is not present)
				% (soz. mit der Spitze mitschwimmen)
				time_to_cork = 19;
				WCdiff = x(newtime,1) - x(newtime,3);
				%if (27.5 <= WCdiff && WCdiff <= 32.5)
				if (WCdiff <= 32.5)
					time_to_cork = 38;
					when_does_C_part = find(x(newtime:stoptime,1)-x(newtime:stoptime,3) > 32.5, 1) - 1;
					if (size(when_does_C_part)(1) == 0)
						when_does_C_part = 0;
						time_to_cork = newtime - stoptime + 1;
					end
					time_to_cork = time_to_cork + when_does_C_part;
				end
				time_to_cork = min(stoptime+1 - newtime, time_to_cork);% Access safety.
				state = "corking";
				
				Dfunction = [1 : time_to_cork]' .* speed_W;
				nc = next_countryside(countrYplace, lastpos);
				if (nc && nc(1,2) < Dfunction(time_to_cork,1) + lastpos)
					last = find(Dfunction .+ lastpos > nc(1,2), 1);
					Dfunction (last:time_to_cork,:) = [];
					time_to_cork = last - 1;
					state = "seekpos";
					reftgt = 30;
				endif
				x(newtime:newtime+time_to_cork - 1, 4) = Dfunction .+ lastpos;
				
				lastpos = lastpos + Dfunction(time_to_cork,1);
				newtime = newtime + time_to_cork; % Neuer Index für den Zugriff auf x.
				continue;
			endif
			
			if (strcmp (state, "corking"))
				
				% Now we have to find out when the end of the convoy reaches the
				% position where C is standing.
				time_to_cnend = find(x(newtime:stoptime,5) + 45 > lastpos, 1);
				if (size(time_to_cnend)(1) == 0)
					time_to_cnend = stoptime+1 - newtime;% Access safety.
				endif
				
				
				Dfunction = zeros(time_to_cnend, 1);
				% D reached the "end" (45 m before) by standing still.
				x(newtime:newtime+time_to_cnend-1,4) = Dfunction .+ lastpos;
				
				lastpos = lastpos + Dfunction(time_to_cnend,1);
				newtime = newtime + time_to_cnend;
				state = "tofront";
				continue;
			endif
			
			
			% tofront and seekpos are similiar, so combine them
			if (strcmp(state, "tofront") || strcmp(state, "seekpos"))
				% Now C proceeds to the front (30 m after W) again.
				
				if (strcmp(state, "tofront"))
					fastspeed = 5.5; % m/s
					tgtpos = 30; % m behind W
					state = "static";
				else
					state = "pendule";
					if (x(newtime-1,1)-lastpos < 600)
						fastspeed = 1.5; % m/s
						tgtpos = 600;
					elseif (x(newtime-1,1)-lastpos > 650)
						fastspeed = 5.5; % m/s
						tgtpos = 600;
					else
						fastspeed = 5.5; % m/s
						tgtpos = 600; % m behind W
					endif
				endif
				% Function of W: posW-30m + 3.5 m/s * t  (point 30 m after W)
				% Function of C: posC     + 5.5 m/s * t
				% Intersection: 
				%   posW-30m + 3.5 m/s * t = posC     + 5.5 m/s * t  | - 3.5 m/s * t
				%   posW - 30 m            = posC     + 2   m/s * t  | - posC |  / 2m/s
				%  (posW - 30 m - posC) / (2m/s)  =               t   
				time_to_front = round( ...
					   1/(fastspeed - speed_W) ...
					 * (x(newtime-1,1)-tgtpos-lastpos)-1);
				
				% Access safety
				time_to_front = min(stoptime+1 - newtime, time_to_front);
				
				
				Dfunction = [1 : time_to_front ]' .* fastspeed;
				nc = next_countryside(countrYplace, lastpos);
				% nc branch duplicated from state = "corking"
				if (nc && nc(1,2) < Dfunction(time_to_front,1) + lastpos && strcmp(state, "static"))
					last = find(Dfunction .+ lastpos > nc(1,2), 1);
					Dfunction (last:time_to_front,:) = [];
					time_to_front = last - 1;
					state = "seekpos";
				endif
				x(newtime:newtime+time_to_front-1,4) = Dfunction .+ lastpos;
				lastpos = lastpos + Dfunction(time_to_front,1);
				newtime = newtime + time_to_front;
				
				% Now C has proceeded to the front. Here the loop begins again.
				continue;
			endif
			
			
			if (strcmp (state, "pendule"))
				% (!) Similar code to D, policy P
				Doffset = 600; % number of meters which D cycle happens behind W
				
				time_to_high = 150; % seconds
				distance_to_high = 50; % meters
				time_to_low = 100; % seconds
				
				minspeed = speed_W - distance_to_high/time_to_high;
				tgt = nc(1,3); % Zum Ende des Landschaftsbereichs
				maxtime = (tgt-lastpos) / minspeed;
				maxtime = min (maxtime, stoptime+1 - newtime);
				
				
				Dfunction1 = [0:time_to_high - 1]' .* distance_to_high/time_to_high;
				Dfunction2 = [time_to_low : -1 : 1]' .* distance_to_high/time_to_low;
				Dfunction_combined = [Dfunction1;Dfunction2];
				
				% Ansatz: wir hängen den Durchlauf nochmal an den Durchlauf an (und
				% verdoppeln dadurch die Größe des Durchlaufs. Das passiert in sehr
				% wenigen Iterationen bis hin zu sehr großen Durchlaufgrößen. 
				% Die durchlaufgröße wächst exponentiell.
				times = ceil (log2(maxtime/size(Dfunction_combined)(1)));
				for i = 1 : times
					Dfunction_combined =[Dfunction_combined;Dfunction_combined];
				endfor
				
				% Dfunction was set up relative to the dynamic position of W.
				x(newtime:newtime+maxtime-1,4) = ...
					   x(newtime:newtime+maxtime-1,1)  ...
					.- Dfunction_combined(1:maxtime,:) ...
					.- Doffset; % D	
				
				ntadd = find (x(newtime:newtime+maxtime-1,4) > tgt, 1);
				if (size(ntadd)(1))
					newtime = newtime + ntadd;
				else
					newtime = newtime + maxtime;
					% We reached end of x.
				endif
				lastpos = x(newtime-1,4);
				state = "tofront";
				
				continue;
			endif
			
			assert(0); % This is unreachable
		endwhile
	
	% Was: Policy corK
	endif % (if -> policy K or P, relevant for D)
	
endfunction



% Helper: next_countryside

function idxetc = next_countryside(arr, start_pos)
	idxetc = find(arr(:,1) > start_pos, 1);
	if (size(idxetc)(1) == 0)
		idxetc = 0;
		return;
	endif
	idxetc(1,2:3) = arr(idxetc,1:2);
endfunction


source("../../Dokument/common/octave-plot.include");


function [] = standard_legend(POS)
	legend("station W (1st)", "station B", "station C", ...
		"station D", "station E", "location", POS);
end


% TODO: Legenden
function [] = plotGenrideDynamics(ctrYarr, a)
	global cW, global cB, global cC, global cD, global cO, global c_landscape
	global l lw n m c ms
	global f, global acc, global Facc
	global medium wide
	
	figure_start(1)
	plot([0,1], [0,0]);
	
	figure_start(2)
	plot([0,1], [0,0]);
	max_height = max(a(5,:).-a(1,:));
	
	figure 1;
	% http://octave.1599824.n4.nabble.com/Reversing-the-x-Axis-When-Plotting-td3949680.html
	% = https://lists.gnu.org/archive/html/help-octave/2011-10/msg00619.html
	set(gca(), "ydir", "reverse");
	
	maxtime = size(a)(1);
	
	stW = 1; stB = 1; stC = 1; stD = 1; stO = 1;
	for i = 1:(size(ctrYarr)(1)+1)
		figure 2;
		if (size(ctrYarr)(1) < i)
			% Ende
			cons=stW:maxtime;
			cons=stB:maxtime;
			cons=stC:maxtime;
			cons=stD:maxtime;
			cons=stO:maxtime;
			plot (cons, a(cons,1)-a(cons,1), lw, 3.0, c,cW);
			plot (cons, a(cons,1)-a(cons,2), lw, 3.0, c,cB);
			plot (cons, a(cons,1)-a(cons,3), lw, 3.0, c,cC);
			plot (cons, a(cons,1)-a(cons,4), lw, 3.0, c,cD);
			plot (cons, a(cons,1)-a(cons,5), lw, 3.0, c,cO);
			hold off
			break;
		else
			cbegin = ctrYarr(i,1);
			enW = find(a(:,1)>cbegin,1); consW = stW:enW;
			enB = find(a(:,2)>cbegin,1); consB = stB:enB;
			enC = find(a(:,3)>cbegin,1); consC = stC:enC;
			enD = find(a(:,4)>cbegin,1); consD = stD:enD;
			enO = find(a(:,5)>cbegin,1); consO = stO:enO;
			UW = plot (consW, a(consW,1)-a(consW,1), lw, 3.0, c, cW);
			UB = plot (consB, a(consB,1)-a(consB,2), lw, 3.0, c, cB);
			UC = plot (consC, a(consC,1)-a(consC,3), lw, 3.0, c, cC);
			UD = plot (consD, a(consD,1)-a(consD,4), lw, 3.0, c, cD);
			UO = plot (consO, a(consO,1)-a(consO,5), lw, 3.0, c, cO);
			cend = ctrYarr(i,2); fullend = size(a)(1);
			stW = find(a(:,1)>cend,1);
			stB = find(a(:,2)>cend,1);
			stC = find(a(:,3)>cend,1);
			stD = find(a(:,4)>cend,1);
			stO = find(a(:,5)>cend,1);
			if stW; consW = enW + 1 : stW - 1; else; consW = enW+1:fullend; endif;
			if stB; consB = enB + 1 : stB - 1; else; consB = enB+1:fullend; endif;
			if stC; consC = enC + 1 : stC - 1; else; consC = enC+1:fullend; endif;
			if stD; consD = enD + 1 : stD - 1; else; consD = enD+1:fullend; endif;
			if stO; consO = enO + 1 : stO - 1; else; consO = enO+1:fullend; endif;
			CW = plot (consW, a(consW,1)-a(consW,1), lw, 1.5, c, cW);
			CB = plot (consB, a(consB,1)-a(consB,2), lw, 1.5, c, cB);
			CC = plot (consC, a(consC,1)-a(consC,3), lw, 1.5, c, cC);
			CD = plot (consD, a(consD,1)-a(consD,4), lw, 1.5, c, cD);
			CO = plot (consO, a(consO,1)-a(consO,5), lw, 1.5, c, cO);
			hold on;
		endif
		
		figure 1;
		xval1 = ctrYarr(i,1);
		if (xval1 > a(size(a)(1),1))
			break;
		endif
		xval2 = ctrYarr(i,2);
		if (xval2 > a(size(a)(1),1))
			xval2 = a(size(a)(1),1) + 200;
		endif
		pArr = [xval1, -50, xval2-xval1, size(a)(1)+100];
		
		HR = rectangle("Position", pArr, lw,1, ...
				"edgecolor", c_landscape, ...
				"facecolor", c_landscape);
		
		hold on;
	endfor
	
	figure 2;
	xlabel ("Time in seconds", f,acc);
	ylabel ("Distance to leading cyclist (station W) in meters", f,acc);
	% Sequence order is completely irrelevant in octave.
	L =legend([UW UB UC UD UO CW CB CC CD CO], ...
			"Station W (urban)", ...
			"Station B (urban)", ...
			"Station C (urban)", ...
			"Station D (urban)", ...
			"Station O (urban)", ...
			"Station W (countryside)", ...
			"Station B (countryside)", ...
			"Station C (countryside)", ...
			"Station D (countryside)", ...
			"Station O (countryside)");
	set(L, "location", "east");
	axis([0, 11400, axis()([3 4])])
	set(gca, "xtick", [0, 1800, 3600, 5400, 7200])
	print_ticks_1_000("x")
	set(gca, "ytick", [0, 200, 400, 600, 800, 900])
	finish_optics(wide)
	print ("-depsc", "distance_to_lead.eps", Facc)
	print ("-dpng", "-r450", "distance_to_lead.png", Facc)
	
	
	figure 1;
	timecnt = 1:size(a)(1);
	HW = plot(a(:,1), timecnt, lw, 0.9);
	HB = plot(a(:,2), timecnt, lw, 0.9);
	HC = plot(a(:,3), timecnt, lw, 0.9);
	HD = plot(a(:,4), timecnt, lw, 0.9);
	HO = plot(a(:,5), timecnt, lw, 0.9);
	HRfake = plot(10000, 500000, l,n, "markerfacecolor", c_landscape, m,'s');
	%annotation('rectangle',[.65 .62 .34, .37],'edgecolor','k','linewidth',1,"facecolor", 'w');
	L =legend([HW, HB, HC, HD, HO, HRfake], {"Station W (1st)", "Station B (changing)", ...
		"Station C (changing)", "Station D (changing)", "Station O (last)", "Countryside"});
	set(L, "location", "north");
	axis([0000, 27000, -5700, 8000])
	xlabel ("Distance from starting point in meters", f,acc);
	ylabel ("Time after leading cyclist (station W) started", f,acc);
	set(gca, "xtick", [0, 5000, 10000, 15000, 20000, 25000])
	print_ticks_1_000("x")
	set(gca, "ytick", [0, 2000, 4000, 6000, 8000])
	print_ticks_1_000("y")
	%LI = line([1,0],[0,2],"color", 'g', "marker", 'x', "parent", L)
	%get(L)
	%legend("toggle")
	%legend("toggle")
	finish_optics(wide)
	%set(L, "layer", "top")
	print ("-depsc", "time_to_lead.eps", Facc)
	print ("-dpng", "-r450", "time_to_lead.png", Facc)
	
	set([HW, HB, HC, HD, HO], lw, 3);
	set (gca, "xtickmode", "auto")
	set (gca, "xticklabelmode", "auto")
	set (gca, "yticklabelmode", "auto")
	set (gca, "ytickmode", "auto")
	axis([1800, 11041, 606, 3147])
	%set(gca, "xgrid", "on");
	%axis([8000, 11500, 2150, 3615])
	print_ticks_1_000("x")
	print_ticks_1_000("y")
	%standard_legend("northeast")
	legend("off")
	L = legend([HW, HB, HC, HD, HO], {"Station W (1st)", ...
		"Station B (changing)", ...
		"Station C (changing)", "Station D (changing)", "Station O (last)"});
	set(L, "location", "northeast");
	finish_optics(medium)
	print ("-depsc", "time_to_lead-detail.eps", Facc)
	print ("-dpng", "-r450", "time_to_lead-detail.png", Facc)
	
	%TODO: set linewidth (5x H, 1x HR)
	%Various Legends (H)
	
	
	
endfunction




