

// © 2018 Jonas Bechtel
// Licensed under the Apache License, Version 2.0; see file <prj>/meta/LICENSE


/*
 * r-map makes following:
 * 
 * - Load original AMR data and assign ideal timestamp
 * - Load received AMR data with timestamp and sequence number
 * - Compare and map the AMR data
 *    -> count how many (equal) packets arrived per original frame
 * - Output following in octave readable format:
 *    * total list of received sequence numbers for each sequence number.
 *    * list of most frequent received sequence number for each sequence number
 *    * for every count: the most frequent received sequence number for each
 *           sequence number
 *    * the first time of arrival for each sequence number
 * 
 */

#ifndef RMAP_C
#define RMAP_C

#define BREAK_AT (7400*50)
#define STATION_NAME 'W'


#include <stdbool.h>
#include <stdint.h>

#ifndef NOCODE

#include <stdio.h>
#include <stdlib.h>
#include <locale.h> // setlocale()

#define NOCODE


#undef RMAP_C
#include "r-map.c"

#include "read-ulog.c"

#undef NOCODE


#endif



#ifdef NOCODE

#include "amrload.c"


struct subseq {
	uint16_t sequence_number;
	uint8_t hit; // Only needed for final stage
	uint64_t toa;
};


struct rmap_frame {
	struct mr475_frame data;
	uint32_t sequence_number;
	uint64_t toa;
#define NUM_SUBSEQ 7
	uint8_t cnt_sub;
	struct subseq sub[NUM_SUBSEQ];
	
};

struct rmap_framegroup {
#define FRAMEGROUP 40
//#define FRAMEGROUP 630   // sqrt(8000*50) = 10 sqrt(4000) = 100 sqrt(40) = 200 sqrt(10) ~= 630
	struct rmap_frame data[FRAMEGROUP];
	uint8_t count;
	struct rmap_framegroup* last;
	struct rmap_framegroup* next;
};

struct rmap {
	struct amrload amrload;
	
//#define MAXFRAMES 8000*50 // reaches surely till 7200 seconds
	//struct mr475_frame[MAXFRAMES];
	
	struct rmap_framegroup sorted;
	
	struct rmap_framegroup* fastlist[256];
	
	struct rmap_framegroup* sorted_ptr[BREAK_AT];
	uint16_t sorted_pos[BREAK_AT];
};

#endif

#define A_FIRST -1
#define EQUAL 0
#define B_FIRST 1
int8_t compare(struct mr475_frame a, struct mr475_frame b)
#ifndef NOCODE
{
	uint8_t pos = sizeof(a.data);
	
	while (pos-- > 0) {
		if (a.data[pos] == b.data[pos])
			continue;
		if (a.data[pos] < b.data[pos])
			return A_FIRST;
		else
			return B_FIRST;
	}
	return EQUAL;
}
#endif
;


// If entry is fresh, new is set. new has to be NULL otherwise
void tick_frame(struct rmap_frame * rf, uint32_t seqno, struct mr475_frame * new)
#ifndef NOCODE
{
	rf->sequence_number = seqno;
	rf->data = *new;
	rf->cnt_sub = 0;
	// TODO: count, pwr, ...?
}
#endif
;


// puts NULL in res if not found.
// returns the # inside framegroup
int find_frame(struct rmap* rmap, struct rmap_framegroup ** res, struct mr475_frame search)
#ifndef NOCODE
{
	
	//struct rmap_framegroup* ptr = &rmap->sorted;
	struct rmap_framegroup* ptr = rmap->fastlist[search.data[11]];
	
	while (ptr->next) {
#ifdef DEBUG_SORT
		printf("ptr->next:%p\n", ptr->next);
#endif
		ptr = ptr->next;
		
		int8_t cres = compare(search, ptr->data[0].data);
		
		if (cres == EQUAL) {
#ifdef DEBUG_SORT
			printf("Already present.\n");
#endif
			*res = ptr;
			return 0;
		}
		
		if (cres == A_FIRST) {
#ifdef DEBUG_SORT
			printf("point reached.\n");
#endif
			if (ptr->last->last)
				ptr = ptr->last;
			break;
		}
		
		if (cres == B_FIRST) {
#ifdef DEBUG_SORT
			printf("searching.\n");
#endif
			continue;
		}
		
	}
	
	if (ptr->last == NULL) {
		*res = NULL;
		
#ifdef DEBUG_SORT
		print_state(rmap);
#endif
		return 0;
	}
	
	
	for (int i = 0; i < ptr->count; i++) {
		
		int8_t cres = compare(search, ptr->data[i].data);
		
#ifdef DEBUG_SORT
		printf("i=%d, ptr->count=%d, res=%s.\n", i, ptr->count, res == EQUAL ? "EQUAL" : res == A_FIRST ? "A_FIRST" : "B_FIRST");
		//print_state(rmap);
#endif
		
		
		
		if (cres == EQUAL) {
#ifdef DEBUG_SORT
			printf("Found.\n");
#endif
			*res = ptr;
			return i;
		}
		
		if (cres == B_FIRST) {
			continue;
		}
		
		if (cres == A_FIRST) {
			//printf("Not found!\n");
			*res = NULL;
			return 0;
		}
	}
	
	
	//printf("Not found! (II)\n");
	*res = NULL;
	return 0;
	
	
}
#endif
;



void insert_frame(struct rmap * rmap, uint32_t seqno, struct mr475_frame new)
#ifndef NOCODE
{
	
	//struct rmap_framegroup* ptr = &rmap->sorted;
	struct rmap_framegroup* ptr = rmap->fastlist[new.data[11]];
	
	while (ptr->next) {
#ifdef DEBUG_SORT
		printf("ptr->next:%p\n", ptr->next);
#endif
		ptr = ptr->next;
		
		int8_t res = compare(new, ptr->data[0].data);
		
		if (res == EQUAL) {
#ifdef DEBUG_SORT
			printf("Already present.\n");
#endif
			tick_frame(&ptr->data[0], seqno, NULL);
			return;
		}
		
		if (res == A_FIRST) {
#ifdef DEBUG_SORT
			printf("point reached.\n");
#endif
			if (ptr->last->last)
				ptr = ptr->last;
			break;
		}
		
		if (res == B_FIRST) {
#ifdef DEBUG_SORT
			printf("searching.\n");
#endif
			continue;
		}
		
	}
	
	if (ptr->last == NULL) {
		//printf("A\n");
		struct rmap_framegroup* rf = malloc(sizeof (struct rmap_framegroup));
		if (rf == NULL) {
			fprintf(stderr, "malloc->rf failed!\n");
		}
		rf->count = 1;
		tick_frame(&rf->data[0], seqno, &new);
		//rmap->sorted_ptr[seqno] = rf;
		
		rf->next = rmap->sorted.next;
		rf->last = &rmap->sorted;
		rmap->sorted.next = rf;
		if (rf->next)
			rf->next->last = rf;
		
#ifdef DEBUG_SORT
		print_state(rmap);
#endif
		return;
	}
	
#define MODE_SEARCH 0
#define MODE_MOVE 1
	//uint8_t mode = MODE_SEARCH;
	int i = 0;
	if (ptr->count > 5) {
		int8_t res = compare(new, ptr->data[ptr->count/2].data);
		
#ifdef DEBUG_SORT
		printf("i=%d, ptr->count=%d, res=%s.\n", ptr->count/2, ptr->count, res == EQUAL ? "EQUAL" : res == A_FIRST ? "A_FIRST" : "B_FIRST");
		//print_state(rmap);
#endif
		if (res == B_FIRST)
			i = ptr->count/2;
	}
	
	for (; i < ptr->count; i++) {
		
		int8_t res = compare(new, ptr->data[i].data);
		
#ifdef DEBUG_SORT
		printf("i=%d, ptr->count=%d, res=%s.\n", i, ptr->count, res == EQUAL ? "EQUAL" : res == A_FIRST ? "A_FIRST" : "B_FIRST");
		//print_state(rmap);
#endif
		
		
		
		if (res == EQUAL) {
#ifdef DEBUG_SORT
			printf("Already present.\n");
#endif
			tick_frame(&ptr->data[i], seqno, NULL);
			//rmap->sorted_ptr[seqno] = ptr;
			return;
		}
		
		if (res == B_FIRST) {
			continue;
		}
		
		if (res == A_FIRST) {
			if (ptr->count > FRAMEGROUP) {
				fprintf(stderr, "Very strange!\n");
				exit (1);
			}
			if (ptr->count == FRAMEGROUP) {
				// New framegroup
				struct rmap_framegroup * rfg = malloc (sizeof(struct rmap_framegroup));
				if (rfg == NULL) {
					fprintf(stderr, "malloc->rfg failed!\n");
				}
				
				if (ptr->next)
					ptr->next->last = rfg;
				rfg->next = ptr->next;
				ptr->next = rfg;
				rfg->last = ptr;
				
				rfg->count = 1;
				if (i < ptr->count-1) {
					rfg->data[0] = ptr->data[ptr->count - 1];
					
					for (int j = ptr->count-1; j > i; j--) {
						ptr->data[j] = ptr->data[j-1];
					}
					tick_frame(&ptr->data[i], seqno, &new);
					//rmap->sorted_ptr[seqno] = ptr;
					
#ifdef DEBUG_SORT
					print_state(rmap);
#endif
					return;
				}
				
				tick_frame(&rfg->data[0], seqno, &new);
				//rmap->sorted_ptr[seqno] = rfg;
				
#ifdef DEBUG_SORT
				print_state(rmap);
#endif
				return;
			}
			
			
			for (int j = ptr->count; j > i; j--) {
				ptr->data[j] = ptr->data[j-1];
			}
			
#ifdef DEBUG_SORT
			printf("Tick %d.\n", i);
#endif
			tick_frame(&ptr->data[i], seqno, &new);
			//rmap->sorted_ptr[seqno] = ptr;
			
			ptr->count++;
			
#ifdef DEBUG_SORT
			print_state(rmap);
#endif
			return;
		}
	}
	
	if (i > FRAMEGROUP) {
		fprintf(stderr, "Very strange (II)!\n");
		exit (1);
	}
	
	if (i == FRAMEGROUP) {
		// New framegroup
		struct rmap_framegroup * rfg = malloc (sizeof(struct rmap_framegroup));
		if (rfg == NULL) {
			fprintf(stderr, "malloc->rfg failed (II)!\n");
		}
		
		if (ptr->next)
			ptr->next->last = rfg;
		rfg->next = ptr->next;
		ptr->next = rfg;
		rfg->last = ptr;
		
		rfg->count = 1;
		
		tick_frame(&rfg->data[0], seqno, &new);
// 		rmap->sorted_ptr[seqno] = rfg;
			
#ifdef DEBUG_SORT
		print_state(rmap);
#endif
		return;
	}
	
	
	ptr->count++;
	
	tick_frame(&ptr->data[i], seqno, &new);
	// rmap->sorted_ptr[seqno] = ptr;
	
#ifdef DEBUG_SORT
	print_state(rmap);
#endif
}
#endif
;


void print_state(struct rmap* rmap)
#ifndef NOCODE
{
	struct rmap_framegroup* ptr = rmap->sorted.next;
	
	
	printf("SORTING STATE\n--------\n");
	
	while (ptr) {
		
		
		int i;
		for (i = 0; i < ptr->count; i++) {
			printf("a|%5x", ptr->data[i].sequence_number);
			
			int k;
			for (k = 0; k < 12; k++) {
				if (!k%4) {
					printf("|");
				}
				
				printf("%02x", ptr->data[i].data.data[k]);
			}
			
			if (ptr->data[i].cnt_sub == 0)
				printf(" - \n");
			else
				printf(" %d \n", ptr->data[i].cnt_sub);
			
		}
		printf("--------\n");
		
		ptr = ptr->next;
	}
	
}
#endif
;

#ifndef NOCODE
#pragma GCC diagnostic ignored "-Wformat"
#endif

#define HUMAN true
#define OCTAVE false
#define ALL 0xFF
void print_time_state(struct rmap * rmap, bool human, uint8_t select, int rows)
#ifndef NOCODE
{
	
	printf(human ? "TIME-BASED STATE for human\n" : "# TIME-BASED STATE for octave\n");
	if (select != ALL)
		printf(human ? "select frames with %d arrivals\n" : "# name: select%d\n", select);
	if (human)
		printf("------------------------\n");
	else {
		printf("# type: matrix\n");
		printf("# rows: %d\n", rows);
		printf("# columns: 4\n");
	}
	
	int i;
	for (i = 0; i < BREAK_AT; i++) {
		uint16_t p = rmap->sorted_pos[i];
		struct rmap_framegroup* ptr = rmap->sorted.next;
		ptr = rmap->sorted_ptr[i];
		
		if (select != ALL && ptr->data[p].cnt_sub != select)
			continue;
		
		printf(human ? "a|%5x|...|: " : " %d ", ptr->data[p].sequence_number);
		
		
		if (ptr->data[p].cnt_sub == 0)
			printf(human ? "- \n" : "1500 0 0\n");
		else {
			int j;
			uint8_t max_idx = 0;
			for (j = 0; j < ptr->data[p].cnt_sub; j++) {
				if (ptr->data[p].sub[j].hit > ptr->data[p].sub[max_idx].hit)
					max_idx = j;
			}
			printf(human ? "%3x(%d), first@ %'lu, " : "%1$d %3$lu ",
			        ptr->data[p].sub[max_idx].sequence_number,
			        ptr->data[p].sub[max_idx].hit,
			        ptr->data[p].sub[max_idx].toa);
			printf(human ? "%d total \n": "%d\n", ptr->data[p].cnt_sub);
		}
		
	}
	
	printf("\n\n");
}
#endif
;




// Assumes that at least one frame is read in.
void update_fastlist(struct rmap* rmap)
#ifndef NOCODE
{
	int i = 0;
	
	struct rmap_framegroup* ptr = rmap->sorted.next;
	struct rmap_framegroup* found = &rmap->sorted;
	
	while (ptr) {
		if (ptr->data[0].data.data[11] >= i) {
			for (int j = i; j <= ptr->data[0].data.data[11]; j++) {
				rmap->fastlist[j] = ptr->last;
			}
			i = ptr->data[0].data.data[11] + 1;
			found = ptr;
		}
		ptr = ptr->next;
	}
	
	
	for (int j = i; j <= 0xFF; j++) {
		rmap->fastlist[j] = found;
	}
	
#ifdef DEBUG_SORT
	printf("fastlist\n--------\n");
	
	for (i = 0; i < 256; i++) {
		if (rmap->fastlist[i]->last == NULL)
			printf("%02x - start\n", i);
		else {
			printf("%02x - %5x", i, rmap->fastlist[i]->data[0].sequence_number);
			
			int k;
			for (k = 0; k < 12; k++) {
				if (!k%4) {
					printf("|");
				}
				
				printf("%02x", rmap->fastlist[i]->data[0].data.data[k]);
			}
			printf("\n");
		}
	}
	printf("\n");
	
	print_state(rmap);
#endif
	
}
#endif
;





#ifndef NOCODE
int main() {
	
	// Thousands' commas and decimal points in any environment.
	setlocale(LC_NUMERIC, "en_US");
	// Remark: you _have_ to call setlocale(LC_NUMERIC, <anything>) in order
	// to get thousands' seperators at all!
	
	
	
	
	struct rmap rmap;
	rmap.sorted.next = NULL;
	rmap.sorted.last = NULL; // this way you recognize rmap.sorted
	rmap.sorted.count = 0; // And it will be ever 0.
	
	bool res = false;
	res = init_amrload(&rmap.amrload);
	
	if (!res) {
		printf("amrload failed!");
		return 1;
	}
	
	for (int i = 0; i < 256; i++) {
		rmap.fastlist[i] = &rmap.sorted;
	}
	
	struct mr475_frame new;
	for (int i = 0; i < BREAK_AT; i++) {
		
		bool res;
		res = read_amr_frame(&rmap.amrload, &new);
		if (!res)
			break;
		insert_frame(&rmap, i, new);
		if (i % 10000 == 1) {
			fprintf(stderr, "Processed %d.\n", i);
		}
		if (i % 500 == 499) {
			// speedup
			update_fastlist(&rmap);
		}
	}
	
	//print_state(&rmap);
	
	
	
	//printf("read_next() = %d -> %3x\n", read_next(&fr, STATION_NAME), fr.sequence_number);
	
	int count_not_found = 0;
	
	int i;
	for (i = 0; i < BREAK_AT; i++) {
		struct rmap_frame fr;
		bool res;
		
		res = read_next(&fr, STATION_NAME);
		if (!res)
			break;
		
		struct	 rmap_framegroup* ptr;
		int pos = find_frame(&rmap, &ptr, fr.data);
		
		
		if (i % 10000 == 1) {
			fprintf(stderr, "Processed %d.\n", i);
		}
		
		
		
		if (ptr == NULL) {
			count_not_found++;
			//printf("Not found [answ:%d]\n", pos);
			continue;
		}
		
		//printf("Found @pos:%d\n", pos);
		//printf("w/seqno:%1$5x(%1$d)\n", ptr->data[pos].sequence_number);
		
		
		uint8_t j = 0xFF;
		while (++j < ptr->data[pos].cnt_sub) {
			if (ptr->data[pos].sub[j].sequence_number != fr.sequence_number)
				continue;
			
			ptr->data[pos].sub[j].hit++;
			break;
		}
		
		
		if (ptr->data[pos].cnt_sub + 1 >= NUM_SUBSEQ) {
			fprintf(stderr, "Error: not enough entries for received sequence numbers!\n");
			continue;
		}
		
		
		j = ptr->data[pos].cnt_sub++;
			
		ptr->data[pos].sub[j].sequence_number = fr.sequence_number;
		ptr->data[pos].sub[j].toa = fr.toa;
		ptr->data[pos].sub[j].hit = 1;
	}
	
	
	printf("# name: count_not_found\n");
	printf("# type: scalar\n");
	printf("%d\n", count_not_found);
	printf("\n\n");
	
	printf("# name: total_received\n");
	printf("# type: scalar\n");
	printf("%d\n", i);
	printf("\n\n");
	
	
	// final sorting ptr
	
	struct rmap_framegroup* ptr = rmap.sorted.next;
	while(ptr) {
		for(int i = 0; i < ptr->count; i++) {
			rmap.sorted_pos[ptr->data[i].sequence_number] = i;
			rmap.sorted_ptr[ptr->data[i].sequence_number] = ptr;
		}
		ptr = ptr->next;
	}
	
	
	
	//print_state(&rmap);
	
	//print_time_state(&rmap, HUMAN, ALL, 5555);
	
	int cnt[7];
	
	for (int t = 0; t < 7; t++) {
		cnt[t] = 0;
	}
	
	for (i = 0; i < BREAK_AT; i++) {
		uint16_t p = rmap.sorted_pos[i];
		struct rmap_framegroup* ptr = rmap.sorted.next;
		ptr = rmap.sorted_ptr[i];
		
		cnt[ptr->data[p].cnt_sub]++;
	}
	
	
		
	for (int t = 0; t < 7; t++) {
		print_time_state(&rmap, OCTAVE, t, cnt[t]);
	}
	
	
	
	
	return 0;
}
#endif








#endif // -ifndef RMAP_C

