

// © 2018 Jonas Bechtel
// Licensed under the Apache License, Version 2.0; see file <prj>/meta/LICENSE



// air.c - Here the different electromagnetic waves get added
//   automagically by physics©.


#ifndef AIR_C
#define AIR_C


#include <stdint.h> // uint32_t
#include <math.h> // float_t
#include <stdlib.h> // malloc()

#include <assert.h>
#include <cassert.h>

#include "sim-common.h"

/*
#ifdef NOCODE
#include "simulator.c"
#else
#define NOCODE
#include "simulator.c"
#undef NOCODE
#endif
*/




struct timing_list {
	int size;
	uint32_t * times; // each in ns, 0 = beginning of this timing list.
	
	// coefficient[i] describes the strength of the 
	float_t * amplitudes;
	float_t * data;
};


// Keep this efficient.
// Input: wb0 (stronger amplitude), wbx (weaker amplitude) with the same duration
//   of both.
// Output: result and "true" on short_wb (if not NULL) whether result is shorter than
//   than initial duration.
// Note that short_wb will not be set to false in any case.
// TODO: Test this function.
struct waveblip combine_waveblips(struct waveblip wb0, struct waveblip wbx, bool* short_wb)
#ifndef NOCODE
{
	// When invoked by aircombine_new(), we are guaranteed that
	//     wbx.amplitude >= 0.1*wb0.amplitude
	
	
	float p0 = wb0.phase;
	float px = wbx.phase;
	px -= p0;
	
	uint32_t freq = wb0.frequency - wbx.frequency;
	int32_t freqdiff = (int32_t)freq;
	if (freqdiff < 0)
		freqdiff = - freqdiff;
	
	// TODO: Bei den Beschreibungen die Glossaranteile aus den Modi-Beschreibungen
	// rausholen und auf die globale Ebene stellen.
	
	
	
	// Wir unterscheiden vier Modi:
	// Modus D: wb0 gewinnt (stärkere Amplitude)
	// Modus C: sehr schnell wechselnde Inteferenzwirkung.
	//          (> 60 Wechsel im Betrachtungszeitraum)
	// Modus B: langsam wechselnde Inteferenzwirkung.
	// Modus A: dauerhaft wirkende Inteferenz
	//          (< 0,032 Wechsel im Betrachtungszeitraum)
	
	// Die Herangehensweise bedeutet, dass nicht immer gleiche Ergebnisse
	// für gleiche Daten erziehlt werden werden, da die Betrachtungszeiträume
	// je nach Situation wechseln können. (Z. B. wenn mehrere gleiche Bits zu
	// einem waveblip zusammengefasst werden.) Allerdings gehe ich davon aus,
	// dass die angegebenen Werte jederzeit ein ordentliches Ergebnis liefern.
	
	
	// .1*pi/.001 = 209.4   ("stable" = not changing more than 0.1)
/*	if (freqdiff < 210) {
		// Mode A: constructive/destructive interference.
		//         which is stable for at least 1.5 ms.
		//         (depending on phase it may be stable for longer time)
		*/
		
			// .1*pi*1000000000 = 314159265
			// We must cast (float) because otherwise there's some floating
			// point exception on small values of wb0.duration.
	if (freqdiff < 314159265/(float)(wb0.duration)) {
		// Mode A: constructive/destructive interference.
		// Quality of mode: overall good
		//        (medium: calculation/simplifications untested)
		//        (good: data formats suitable for result output)
		
		// The amplitude level of the summed up signal will change only by 
		// factor 0.1 of the total amplitude.
		// As sin(0.1) ~= 0.1 this means that in a 6.3 cycle (2 pi cycle) 
		// the inteference phase change in the given time window will only
		// cover .1/(2*pi) = 0.016 of an inteference cycle. As one inteference
		// cycle contains two changes (high to low and low to high) we say
		// that 2*0.016 = 0.032 changes happen each cycle.
		
		// Abschätzung der Phasen und Amplituden:
		// wb0.a=1, .p=0   wbx.a=1,  .p=pi/2    ->   .a=1.414, .p=pi/4
		// wb0.a=1, .p=0   wbx.a=1,  .p=pi      ->   .a=0,     .p=*
		// wb0.a=1, .p=0   wbx.a=.5, .p=pi/2    ->   .a=1.112, .p=0.463646
			// (http://www.wolframalpha.com/input/?i=cos(x)%2B.5cos(x-pi%2F2) 
			// -> alternate form http://www.wolframalpha.com/input/?i=1.11803+sin(1.10715+%2B+x)&lk=1&rawformassumption=%22ClashPrefs%22+-%3E+%7B%22Math%22%7D
			// -> manual http://www.wolframalpha.com/input/?i=1.11803+cos(1.10715-pi%2F2+%2B+x)&lk=1&rawformassumption=%22ClashPrefs%22+-%3E+%7B%22Math%22%7D
			// )
		
		
		// This is not very precise but something we can handle somehow:
		p0 = (wb0.amplitude * wb0.phase + wbx.amplitude + wbx.phase)/(wb0.amplitude+wbx.amplitude);
		
		wb0.amplitude = wb0.amplitude - wbx.amplitude + 2*wb0.amplitude*cosf(-px/2);
		wb0.frequency = (wb0.frequency + wbx.frequency) / 2;
		wb0.phase = p0;
		if (wb0.amplitude < 0) {
			wb0.amplitude = -wb0.amplitude;
			wb0.phase += 128; // turn around by pi. (pi=128)
		}
		// TODO: Test. There may be many cases with wrong results.
	}
	else if ((uint64_t)freqdiff * (uint64_t)wb0.duration > 60000000000) {
		
		// Mode C: Schwebung, d. h. ein ständiges Hin- und Herwechseln zwischen
		// konstruktiver und destruktiver Inteferenz. In den destruktiven Zeiten
		// herrscht "Funkstille", in den konstruktiven Zeiten "funktionieren"
		// beide Frequenzen. [Genauer gesagt wird die stärkere Welle besser in
		// der resultierenden Welle repräsentiert. Die schwächere Welle hat 
		// teils erhebliche Phasendifferenzen zu der resultierenden Welle. Für
		// beide Quellwellen gilt: ihre Phase passt immer bei der konstruktivsten
		// Inteferenzphase am besten zusammen.]
		
		// Im Endeffekt ist also die Leistung reduziert, die Endfrequenz kann
		// aber als recht stabil und technisch gut zu verarbeiten bezeichnet
		// werden.
		
		
		// > 60 Wechsel im Betrachtungszeitraum: d. h.:
		//    freq_diff * t_win > 60
		//    
		
		// Quality of mode: overall good
		//        (medium: calculation/simplifications untested)
		//        (good: data formats suitable for result output)
		
		
		// Wild guess. Die Resultierende ist eine Mischung aus den Frequenzen;
		// in den destruktiven Phasen sind aber nicht so einfach die
		// Trägerfrequenzbögen zu zählen.
		wb0.frequency = 
				(wb0.amplitude * wb0.frequency + wbx.amplitude * wbx.frequency)
						/ (wb0.amplitude + wbx.amplitude);
		
		// Die Phase ist nicht determinierbar. Wir nutzen 8 von 31 Zufalls-Bits.
		wb0.phase = rand() & 0xFF; 
		wb0.amplitude = ((wb0.amplitude + wbx.amplitude)/sqrt(2))/2;
	}
	else {
		// Mode B: few times changing inteference.
		// Worst mode for calculation.
		
		// Quality of mode: bad
		//        (medium: calculation/simplifications untested)
		//        (bad: data formats suitable for result output)
		
		// We extend the maximum acceptable amplitude factor to 0.3
		// In other words: if the small amplitude is not more than 0.3 of the
		// strong amplitude, then the strong wave will be copied and downfactored
		// and the weak wave will be skipped.
		
		// In the other cases we construct a short wave blip only containing a
		// small fraction of data but this in a correct way.
		
		
		if (wbx.amplitude < 0.3 * wb0.amplitude) {
			// Factor extended case.
			wb0.amplitude -= wbx.amplitude;
			// Keeping phase and frequency.
		}
		else {
			// First arc only
			// This is the worst case for the simulation.
			
			// freq_diff * t_win !!!= .3 -> t_win = .3/freq_diff
			wbx.duration = 300000000/freqdiff; // temporary
			assert(wbx.duration < wb0.duration);
			wb0.duration = wbx.duration;
			wb0.frequency = 
				(wb0.amplitude * wb0.frequency + wbx.amplitude * wbx.frequency)
						/ (wb0.amplitude + wbx.amplitude);
			
			// Phase is undeterminable. Wir use 8 out of 31 random bits.
			wb0.phase = rand() & 0xFF;
			
			*short_wb = true;
		}
		
	}
	
	return wb0;
}
#endif // ifndef NOCODE
;

// Besides adding waves aircombine has the side effect of accurate 
// time segmenting the waves for output.
struct wave_seq_segment aircombine_new(struct wave_seq_segment* rxbuf, const int rxbufbreak, uint64_t simtime, uint64_t breaktime, uint64_t lazy_wavebreak)
#ifndef NOCODE
{
	uint64_t nexttime;
	nexttime = 0x5555555555555555; // "-Werror=maybe-uninitialized"
	int i;
	struct wave_seq_segment wss;
	
#ifdef DEBUG_AIR
	printf("aircombine_new(): simtime=%'lu, breaktime=%'lu\n", simtime, breaktime);
#endif
	
#define AIRCBN_NUM_RXBUF 8
	assert(AIRCBN_NUM_RXBUF > rxbufbreak); // assert(1) -> Programmabbruch
	
	// nexttime_ac
	uint8_t simtime_active; // Or uint16_t, uint32_t, uint64_t, depending on highest expected rxbufbreak.
	uint8_t nexttime_active;
	uint8_t cnt = 0;
	CASSERT(sizeof(simtime_active)*8 >= AIRCBN_NUM_RXBUF, air_c__Choose_uint64_t_or_be_creative); 
	CASSERT(sizeof(nexttime_active)*8 >= AIRCBN_NUM_RXBUF, air_c__Choose_uint64_t_or_be_creative); 
	CASSERT(sizeof(cnt)*8 >= AIRCBN_NUM_RXBUF, air_c__Choose_uint64_t_or_be_creative); 
	
	static uint8_t memerror_cnt;
	
	
	
	// TODO: improve performance by predicting synchronization?
	//       but the simulator is not designed for adding a beat to the waves
	//         because this would disclose the data structure and thus make 
	//         modulation easier.
	
	// At first find start time.
	simtime_active = 0;
	for (i = 0; i < rxbufbreak; i++) {
		if (!rxbuf[i].waves) continue;
#ifdef DEBUG_AIR
		printf("rxbuf[%d].starttime=%lu, rxbuf[%d].airinfo.dist=%u*100ns.\n", i, rxbuf[i].starttime, i, rxbuf[i].airinfo.dist);
#endif
		
		if (rxbuf[i].starttime < breaktime) {
			// Because airinfo handling is ambigous.
			assert(rxbuf[i].airinfo.dist == 0);
		}
		
		if (!simtime_active) {
			if (rxbuf[i].starttime < breaktime) {
				nexttime = rxbuf[i].starttime; // using nexttime temporary here!
				simtime_active = 1 << i;
			}
		}
		else {
			if (rxbuf[i].starttime == nexttime) {
				simtime_active |= 1 << i;
			}
			else if (rxbuf[i].starttime < nexttime) {
				nexttime = rxbuf[i].starttime; // using nexttime temporary here!
				simtime_active = 1 << i;
			}
		}
	}
	
	if (!simtime_active) {
#ifdef DEBUG_AIR
		printf("air.c:aircombine_new(): WARNING: no wave in time window!\n");
#endif
		// There's absolutely no wave in the expected time window.
		// This happens if some reflection vector is used -> only one station
		// sees the reflection, the others execute this if clause.
		// Or if the simulator has errors.
		
		wss.waves = NULL;
		return wss;
	}
	//simtime = 0;
	
	i = 0;
	// Find the relevant wave sequence
	while (!((simtime_active >> i) & 1)) {
		i++;
	}
	//printf ("First relevant wave sequence: %d  (simtime %lu, nexttime %lu).\n",
	//	i, simtime, nexttime);
	
	/*if (simtime_active - (1 << i) == 0) {
		printf("aircombine_new(): nothing to do, only one input.\n");
		wss = rxbuf[i];
		rxbuf[i].waves = NULL;
		// waves->refcount remains same as we create a new ref by returning wss.
		return wss;
	}*/
	
	assert (nexttime == rxbuf[i].starttime); // internal
	
#ifdef DEBUG_AIR_DETAIL
	printf("rxbuf[%d].waves->capacity = %d\n", i, rxbuf[i].waves->capacity);
#endif
	
	int capacity_inc = 0;
	cnt = 0;
	
	// Memory management.
	for(i=0; i < AIRCBN_NUM_RXBUF; i++) {
		if (!((simtime_active >> i) & 1))
			continue;
		
		capacity_inc += rxbuf[i].waves->capacity;
		cnt++;
	}
	
	// +1 ensures at least 1.
	capacity_inc = ((capacity_inc / (3*cnt)) + 1);
	
	
	// Could take any other rxbuf than i as prototype.
	struct wave_sequence * waveout = malloc(sizeof(struct wave_sequence));
	if (!waveout) {
		if (memerror_cnt == 0)
			fprintf(stderr, "air.c:aircombine_new(): ERROR: malloc() fails! "
					"(You are out of memory)\n");
		else if (memerror_cnt == 1)
			fprintf(stderr, "air.c:aircombine_new(): ERROR: malloc() fails; "
					"this won't be reported any more!");
		if (memerror_cnt >= 0 && memerror_cnt <= 5)
			memerror_cnt++;
		
		wss.waves = NULL;
		return wss;
	}
	//printf("waveout= %p \n", waveout);
	waveout->refcount = 0;
	waveout->capacity = 1;
	waveout->duration = 0;
	waveout->skip = 0;
	int waveout_used_seqs = 0;
	
	
	wss.starttime = nexttime;
	wss.airinfo.damp = 1;
	wss.airinfo.dist = 0;
	
	assert(nexttime >= simtime); // What a helpful assert!
	
#ifdef DEBUG_AIR
	if (nexttime - simtime)
		printf("Skipping %lu ns of time.\n", nexttime - simtime);
#endif
	simtime = nexttime;
	
	while (true) {
		nexttime_active = 0;
		for (i = 0; i < rxbufbreak; i++) {
			if (!rxbuf[i].waves) continue;
			
			// Don't need this:
			//if (is_waveblip_invalid(rxbuf[i].first_element)) continue;
			
			
			uint64_t wtime;
			if ((simtime_active >> i) & 1) {
				wtime = rxbuf[i].starttime + rxbuf[i].first_element.duration;
			}
			else {
				wtime = rxbuf[i].starttime;
			}
			
			// New rule: first_element is always valid, so we don't have to check.
			if (!nexttime_active) {
				if (wtime <= breaktime) {
					nexttime = wtime;
					nexttime_active = 1 << i;
				}
				else if (wtime <= breaktime + lazy_wavebreak) {
					// Allowing the waveblip to keep complete.
					nexttime = wtime;
					nexttime_active = 1 << i;
				}
				else {
					nexttime = breaktime;
					nexttime_active = 1 << i;
				}
			}
			else {
				if (wtime == nexttime) {
					nexttime_active |= 1 << i;
				}
				else if (wtime < nexttime) {
					nexttime = wtime; // using nexttime temporary here!
					nexttime_active = 1 << i;
				}
			}
			
		}
#ifdef DEBUG_AIR_DETAIL
		printf ("simtime_active: 0x%x (simtime=%lu), nexttime_active 0x%x (nexttime=%lu).\n", simtime_active, simtime, nexttime_active, nexttime);
#endif
		
		if (!nexttime_active) {
#ifdef DEBUG_AIR
			printf("air.c:aircombine_new(): finished time window (wave ended).\n");
#endif
			// There's absolutely no wave in the expected time window.
			break;
		}
		// Now between simtime and nexttime there's data to merge.
		
		
		i = 0;
		
		
		struct waveblip wb[AIRCBN_NUM_RXBUF];
		cnt = 0;
		
		// Find the relevant wave sequence
		while (i < AIRCBN_NUM_RXBUF) {
			if ((simtime_active >> i) & 1) {
				
				
				if(/*rxbuf[i].waves == NULL ||*/ rxbuf[i].segment_duration == 0) {
#ifdef DEBUG_AIR_DETAIL
					printf("SEGMENT FINISHED.\n");
#endif
					clean_wave_seq_segment(&rxbuf[i]);
					simtime_active ^= 1 << i;
					continue;
				}
				
				assert2 (rxbuf[i].waves != NULL, "This shouldn't happen.");
				
				// prepare waveblips.
				struct waveblip first_element = rxbuf[i].first_element;
				
				if (is_waveblip_invalid(first_element)) {
					//printf ("duration = %lu, refcount = %d, capacity = %d.\n",
					//		rxbuf[i].waves->duration, rxbuf[i].waves->refcount,
					//		rxbuf[i].waves->capacity);
					fflush (stdout);
				}
				
				assert (!is_waveblip_invalid(first_element));
				
				
#ifdef DEBUG_AIR_DETAIL
				printf("rxbuf[%d].starttime = %lu,   rxbuf[%d].first_element.duration = %u.\n", i, rxbuf[i].starttime, i, first_element.duration);
#endif
				
				
				// From read_ride.c: Damp really means damp. So if damp == 2
				// then a wave sent out with  1 mW  will be received  @ 0.5 mW.
				first_element.amplitude /= rxbuf[i].airinfo.damp;
#ifdef DEBUG_AIR_DETAIL
				printf("rxbuf damp: %f\n", rxbuf[i].airinfo.damp);
#endif
				
				
				if (rxbuf[i].starttime + first_element.duration > nexttime) {
					// The element in wave_sequence takes more time than 
					// the slice we process here -> the remaining time must
					// be set to the remaining waveblip.
					
					// Cut
					
					// TODO: Test this thoroughly:
					rxbuf[i].first_element.duration -= nexttime - simtime;
					rxbuf[i].first_element.phase += ((uint64_t)256*(uint64_t)(nexttime - simtime))/(uint64_t)first_element.frequency;
					
					first_element.duration = nexttime - simtime;
					
#ifdef DEBUG_AIR_DETAIL
					printf("Cutting to duration %u .\n",
						   first_element.duration);
#endif
				}
				else {
#ifdef DEBUG_AIR_DETAIL
					printf("Starttime: %lu, [nexttime: %lu], duration: %u, next: %u\n", rxbuf[i].starttime, nexttime, first_element.duration, rxbuf[i].num_second_element);
#endif
					rxbuf[i].first_element
						= rxbuf[i].waves->seq[rxbuf[i].num_second_element];
					rxbuf[i].num_second_element++;
					
					if (rxbuf[i].num_second_element > rxbuf[i].waves->capacity) {
						// break condition:
						// - either amplitude is already NAN
						// - or the capacity is exceeded, then we have to invalidate it.
						
						// harmonizing break -> communicating NAN.
						rxbuf[i].first_element.amplitude = NAN;
					}
					
						// This catches the NAN:
					if (is_waveblip_invalid(rxbuf[i].first_element)) {
						
#ifdef DEBUG_AIR_DETAIL
						printf("SEGMENT FINISHED. (II)\n");
#endif
						// For communication in outer loop
						clean_wave_seq_segment(&rxbuf[i]); // rxbuf[i].waves = NULL;
						
						// For communication in inner loop
						rxbuf[i].segment_duration = 0; 
					}
#ifdef DEBUG_AIR_DETAIL
					// Da ändert sich gar nichts.
					//printf("*starttime: %llu, [nexttime: %llu], duration: %u\n", rxbuf[i].starttime, nexttime, first_element.duration);
#endif
				}
				
				if (rxbuf[i].waves) {
					// rxbuf[i].starttime == simtime for all relevant i.
					assert(rxbuf[i].segment_duration >= nexttime - simtime);
					
					rxbuf[i].segment_duration -= nexttime - simtime;
					rxbuf[i].starttime = nexttime;
				}
				
				wb[cnt] = first_element;
				cnt++;
			} // if simtime_active[...]
			i++;
		}
		
		if (!cnt) {
#ifdef DEBUG_AIR
			printf("aircombine_new(): All waves ended.\n");
#endif
			break;
		}
		
		// Now sort the waveblips (~bubblesort) [I guess this sort is efficient
		//  for very small numbers of joining waves.]
		int c = 0;
		bool change = true;
		while (change) {
			// Forward and backward ...
			change = false;
			while (c < cnt - 1) {
				assert (c+1 < AIRCBN_NUM_RXBUF); // for gcc to be silent.
				if (wb[c].amplitude < wb[c+1].amplitude)  {
					struct waveblip tmp = wb[c];
					wb[c] = wb[c+1];
					wb[c+1] = tmp;
					change = true;
				}
				c++;
				//printf("c=%d, change=%b.\n", c, change);
			}
			if (!change)
				break;
			
			change = false;
			/*while (c >= 0) { Old version.
				if (wb[c].amplitude < wb[c+1].amplitude)  {
					struct waveblip tmp = wb[c];
					wb[c] = wb[c+1];
					wb[c+1] = tmp;
					change = true;
				}
				c--;
				//printf("c=%d, change=%b.\n", c, change);
			}*/
			
			while (c > 0) {
				if (wb[c-1].amplitude < wb[c].amplitude)  {
					struct waveblip tmp = wb[c-1];
					wb[c-1] = wb[c];
					wb[c] = tmp;
					change = true;
				}
				c--;
				//printf("c=%d, change=%b.\n", c, change);
			}
		}
		
		
		bool short_wb = false;
		
#ifdef DEBUG_AIR_DETAIL
		printf ("wb[0].duration = %'u.\n", wb[0].duration);
#endif
		wb[0].duration = nexttime - simtime; // Needed for mode choosing
		
		c = 1;
		while (c < cnt) {
			// HERE we get mathematical:
			
			
			if(wb[c].amplitude < 0.1*wb[0].amplitude)
				break;
			
			wb[0] = combine_waveblips(wb[0], wb[c], &short_wb);
			
			c++;
		}
#ifdef DEBUG_AIR_DETAIL
		printf ("wb[0].duration = %'u.\n", wb[0].duration);
#endif
		
		
		if (waveout->capacity <= waveout_used_seqs + (short_wb?1:0) + 1) {
			waveout->capacity += capacity_inc;
			//printf ("realloc().\n");
			struct wave_sequence * wtmp = 
					realloc(waveout, sizeof(struct wave_sequence)
							+ ((waveout->capacity-1)*sizeof(struct waveblip)));
			if (!wtmp) {
				if (memerror_cnt == 0)
					fprintf(stderr, "air.c:aircombine_new(): ERROR: realloc() "
							"fails! (You are out of memory)\n");
				else if (memerror_cnt == 1)
					fprintf(stderr, "air.c:aircombine_new(): ERROR: realloc() "
							"fails; this won't be reported any more!");
				if (memerror_cnt >= 0 && memerror_cnt <= 5)
					memerror_cnt++;
				continue;
			}
			
			//printf("wtmp= %p \n", wtmp);
			// assert (wtmp == waveout); // This will fail!
			waveout = wtmp;
		}
		
		if (short_wb) {
#ifdef DEBUG_AIR_DETAIL
			printf("SHORT: %d.\n", waveout_used_seqs);
#endif
			wb[1].amplitude = 0;
			wb[1].duration = nexttime - simtime - wb[0].duration;
		}
		else
			wb[0].duration     = nexttime - simtime;
		waveout->duration += nexttime - simtime;
		waveout->seq[waveout_used_seqs] = wb[0];
		waveout_used_seqs++;
		if (short_wb) {
			waveout->seq[waveout_used_seqs] = wb[1];
			waveout_used_seqs++;
		}
		
		
		
		simtime_active |= nexttime_active; // It was a bug to use = instead of |=
		simtime = nexttime;
		
		if (simtime == breaktime) {
#ifdef DEBUG_AIR
			printf("air.c:aircombine_new(): finished time window (keeping rest of wave(s)).\n");
#endif
			// There's absolutely no wave in the expected time window.
			break;
		}
	}
	
	waveout->seq[waveout_used_seqs].amplitude = NAN;
	
	wss.waves = waveout; waveout->refcount++;
	wss.segment_duration = waveout->duration;
	wss.first_element = waveout->seq[0];
	
#ifdef DEBUG_AIR
	printf("aircombine_new(): finished. capacity: %d, used: %d\n", wss.waves->capacity, waveout_used_seqs);
	if (is_waveblip_invalid(wss.first_element)) {
		printf ("waveout->duration = %lu, wavout->refcount = %d, waveout->capacity = %d.\n", waveout->duration, waveout->refcount, waveout->capacity);
		fflush (stdout);
	}
#endif
	wss.num_second_element = 1;
	return wss;
	
	
}
#endif // ifndef NOCODE
;


// Note that 'in' is an _array_  of size 'sz_in'
struct timing_list aircombine_old(int sz_in, struct timing_list * in)
#ifndef NOCODE
{
	struct timing_list ret;
	//if (sz_in == 0) {
		ret.size = 0;
		return ret;
	//}
	/*if (sz_in == 1) {
		ret.size = in[0].size;
		ret.times = in[0].times;
		ret.amplitudes = in[0].amplitudes;
		ret.data = in[0].data;
		return ret;
	}*/
	
	
	// Calculation is not easy for two reasons:
	// 	- Big amounts of data. (Every bit is respected with high timing
	//    precision and multiple bitstreams get combined)
	//  - Determination when a destructive and when a constructive 
	//    inteference is happening. This is dependend on the actual
	//    tranceiver frequency inaccuracy and timing chance (so what's
	//    the "release phase" after one, assuming that the zero is the
	//    normal case.)
	
	// Remark: when calculating multi-channel signals, the inteference
	// in the simulation is the same over time -> inteference is always
	// the same and could be respected by a simple factor.  When calculating
	// signals from multiple transmitters, the factor changes more often
	// (I think) and all the computational effort we do here has a sense.
	
	// So some assumptions and statements for the [statistical] considerations:
	//  * sums of trignonometric functions:
	//			sin(x) + sin(y) = 2 * sin((x+y)/2) * cos((x-y)/2)
	//  * the maximum frequency tolerance is +-1302 Hz. This value is derived
	//			from the future considerations of ITU SM.1045-1 (1997).
	//			[See text file "Frequency Tolerance"]
	//  * -> the frequency offset difference between any two tranceivers is
	//          always 1000 Hz, no matter if this is mathematically impossible.
	//  * to use the full channel bandwidth (25 kHz) I guess the tranceivers
	//			have a frequency deviation of 15 kHz.
	//  * lowest bitrate is 1.200 kbit/s (DEFINITIONS.txt) -> longest 
	//          bit time is 1/(1200kbit/s) = .83 ms, we don't have to consider
	//          longer timespans.
	
	// The advanced formula for the final sum is following:
	// IN: signal 1 @Frequency f1 with Amplitude a1 and Phase p1
	// IN: signal 2 @Frequency f2 with Amplitude a2 and Phase p2
	// O. B. d. A. (english?) we can replace amplitudes and phases:
	// IN: signal 1 @Frequency f1 with Amplitude 1    and Phase p3
	// IN: signal 2 @Frequency f2 with Amplitude 1+a3 and Phase 0
	//       (p3 = p1 - p2, a3 = a2-a1)
	//       (Phase measured in "turnarounds" where 2pi = 1 turnaround.)
	//       (a3 is very high when f2 is stronger, very low when
	//		   f2 is weaker and ~zero when f1 and f2 are similar)
	// 
	// R = a1*cos(f1*t + p3) + a2*cos(f2*t)
	//   = a3*cos(f2*t) + 2*a1*cos((f1*t+p3+f2*t)/2)*cos((f1*t+p3-f2*t)/2)
	//             | As p3 is rather random, p4 = p3/2 will be random as well.
	// R = a3*cos(f2*t) + 2*a1*cos((f1+f2)*t/2+p4)*cos((f1-f2)*t/2+p4)
	
	
	// ACHTUNG, FALSCH (fehlerhafte Annahme cos(x) = e^(jx))
	// R = a1*e^(j*(f1*t+p3))                     + a2*e^(j*(f2*t))
	//   = a1*e^(j*f2*t)*e^(j*p3)*e^(j*(f1-f2)*t) + a2*e^(j*(f2*t))
	//   = e^(j*f2*t)*(a2   +a1*e^(j*p3)*e^(j*(f1-f2)*t))
	
	// RICHTIG (cos(x) = 1/2*(e^(ix) + e^(-ix)))
	// R = .5*a1*(e^(j*(f1*t+p3))+e^(-j*(f1*t+p3)))  + .5*a2*(e^(j*f2*t)+e^(-j*f2*t))
	//   = .5*a1*(e^(j*f2*t)*e^(j*p3)*e^(j*(f1-f2)*t)+e^(-j*f2*t)*e^(-j*p3)*e^(-j*(f1-f2)*t))
	//               + .5*a2*(e^(j*(f2*t))+e^(-j*(f2*t)))
	//   =     .5*e^( j*f2*t)*(a2   +a1*e^(j*p3)*e^(j*(f1-f2)*t))
	//       + .5*e^(-j*f2*t)*(a2   +a1*e^(-j*p3)*e^(-j*(f1-f2)*t))
	
	
	// Ableitung finden: 
	// R'=    .5*j*f2*e^(j*f2*t)*(a2+a1*e^(j*p3)*e^(j*(f1-f2)*t))
	//      + .5*e^(j*f2*t)*(a1*e^(j*p3)*j*(f1-f2)*e^(j*(f1-f2)*t))
	//      - .5*j*f2*e^(-j*f2*t)*(a2+a1*e^(-j*p3)*e^(-j*(f1-f2)*t))
	//      - .5*e^(-j*f2*t)*(a1*e^(-j*p3)*j*(f1-f2)*e^(-j*(f1-f2)*t))
	//   =    .5*j*f2*a2*(e^(j*f2*t)-e^(-j*f2*t))
	//      + .5*j*a1*[ f2*e^(j*p3)*e^(j*f2*t)*e^(j*(f1-f2)*t)
	//                 +(f1-f2)*e^(j*p3)*e^(j*f2*t)*e^(j*(f1-f2)*t)
	//                 -f2*e^(-j*p3)*e^(-j*f2*t)*e^(-j*(f1-f2)*t))
	//                 -(f1-f2)*e^(-jp3)*e^(-j*f2*t)*e^(-j*(f1-f2)*t) ]
	//   =    f2*a2*sin(f2*t)
	//      + .5*j*a1*[ f1*e^(j*f1*t+j*p3) - f1*e^(-j*f1*t-j*p3) ]
	//   =    f2*a2*sin(f2*t) + f1*a1*sin(f1*t+p3)
	
	// davon die Nullstellen finden:
	//  0 =    f2*a2*sin(f2*t) + f1*a1*sin(f1*t+p3)
//	//  sin(f2*t) = - f1/f2*a1/a2*sin(f1*t+p3)                | arcsin (TODO: Boundaries)
//	//      f2*t  = arcsin (-f1/f2*a1/a2*sin(f1*t+p3)) 
//	//         t  = arcsin (-f1/f2*a1/a2*sin(f1*t+p3)) / f2
	//  sin(f1*t+p3) =        - f2/f1*a2/a1*sin(f2*t)         | arcsin (TODO: Boundaries)
	//      f1*t+p3  = arcsin(- f2/f1*a2/a1*sin(f2*t))
	//           p3  = arcsin(- f2/f1*a2/a1*sin(f2*t)) - f1*t
	// P3(t) = Phase, die gewählt werden muss, damit die Funktion für
	//      ein gegebenes t maximal ist.
	
	// Alternativ:
	// 
	//  0 =    f2*a2*sin(f2*t) + f1*a1*sin(f1*t+p3)
	//  0 =    f2*a2*.5/j*(e^(j*f2*t)-e^(-j*f2*t)) + f1*a1*.5/j*(e^(j*f1*t+j*p3)-e^(-j*f2*t-j*p3))
	//  0 =    f2*a2*     (e^(j*f2*t)-e^(-j*f2*t)) + f1*a1*     (e^(j*f1*t+j*p3)-e^(-j*f2*t-j*p3)) | *e^(j*p3)
	//  0 = e^(2j*p3)*f1*a1*e^(j*f1*t)   + e^(j*p3)*f2*a2*(e^(j*f2*t)-e^(-j*f2*t)) - f1*a1*e^(-j*f2*t)
	// ... pq-Formel ...
	// e^(j*p3) = .5 * [-f2*a2*(e^(j*f2*t)-e^(-j*f2*t))
	//				+-sqrt{(-f2*a2*(e^(j*f2*t)-e^(-j*f2*t)))^2-4*f1*a1*e^(j*f1*t)*f1*a1*e^(-j*f2*t)}
	//              ] / (f1*a1*e^(j*f1*t))
	//          = .5 * [-f2*a2*(e^(j*f2*t)-e^(-j*f2*t))
	//				+-sqrt{(f2*a2)^2*(2*j*sin(f2*t))^2        -4*(f1*a1)^2*e^(j*(f1-f2)*t)}
	//              ] / (f1*a1*e^(j*f1*t))
	//          = .5 * [-f2*a2*(e^(j*f2*t)-e^(-j*f2*t))
	//				+-2j*sqrt{(f2*a2)^2*sin(f2*t)^2            + (f1*a1)^2*e^(j*(f1-f2)*t)}
	//              ] / (f1*a1*e^(j*f1*t))
	
	// P3 in die Ausgangsfunktion einsetzen:
	// Rh = a1*cos(f1*t + P3(t)) + a2*cos(f2*t)
	//    = a1*cos(arcsin(-f2/f1*a2/a1*sin(f2*t))) + a2*cos(f2*t)
	
	// Alternativ:
	// Rh=   .5*a1*(e^(j*(f1*t+p3))+e^(-j*(f1*t+p3)))  + .5*a2*(e^(j*f2*t)+e^(-j*f2*t))
	//   =   .5*a2*(e^(j*f2*t)+e^(-j*f2*t)) + .5*a1*e^(j*f1*t)* .5 * [-f2*a2*(e^(j*f2*t)-e^(-j*f2*t))
	//				+-2j*sqrt{(f2*a2)^2*sin(f2*t)^2            + (f1*a1)^2*e^(j*(f1-f2)*t)}
	//              ] / (f1*a1*e^(j*f1*t))
	//         + .5*a1*e^(-j*f1*t)* (f1*a1*e^(j*f1*t))
	//				/ [-f2*a2*(e^(j*f2*t)-e^(-j*f2*t))
	//				+-2j*sqrt{(f2*a2)^2*sin(f2*t)^2            + (f1*a1)^2*e^(j*(f1-f2)*t)}
	//              ]
	//   =   .5*a2*(e^(j*f2*t)+e^(-j*f2*t)) + .25/f1 * [-f2*a2*(e^(j*f2*t)-e^(-j*f2*t))
	//				+-2j*sqrt{(f2*a2)^2*sin(f2*t)^2            + (f1*a1)^2*e^(j*(f1-f2)*t)}
	//              ]
	//         + a1^2*f1
	//				/ [-f2*a2*(e^(j*f2*t)-e^(-j*f2*t))
	//				+-2j*sqrt{(f2*a2)^2*sin(f2*t)^2            + (f1*a1)^2*e^(j*(f1-f2)*t)}
	//              ]
	// (mit +sqrt wird die obere Seite abgebildet, mit -sqrt die untere.)
	// (Oder so. Eigentlich wird gar keine Seite abgebildet, denn die 
	//  Rechnung hat so nicht funktioniert.)
	
	
	// Pre-consideration: 
/*	We will calculate a sum of the waves. The source frequencies are
	most likely somewhat different (500 Hz - 15500 Hz). In case the
	source waves had different amplitudes, there remains another term
	which has another different frequencies whose differences are 
	half of the primary difference (250 Hz -  7750 Hz). As we will
	find out in the following paragraphs, we need to boil it down
	this way until the differential frequencies fall below a fraction
	of the standard bitrate.
	
	
	
	
	When is a frequency difference based addition*/
	//   behaviour same and when it is different? We set the border at
	//   1/sqrt(2) = 0.71 for the strength which the envelope function
	//   (cos) performs. 
	//   But there happens to be a cave-eat: if the starting phase is
	//   at about pi/2, the cos function is sign-changing and a strength
	//   factor doesn't make sense. This affects computation as well as 
	//   reality - the electronics will get puzzled by the rapidly sinking
	//   gain. So we set the resulting power to zero in the [1,3]*p/2
	//   places.
	//   When does the power sinks below 0.71? The time span should be
	//   enough for one Bit (max. ---.83--- ms). UPDATE from below:
	//   in order to allow frequency differences of up to 15,500 Hz
	//   with an knock-out-level @.25 we use a time span of 85 ns.
	//   (~11,8 kbit/s)
	//   
	//   For cos(0 ... pi/2)   [and cos(pi ... 3*pi/2)]
	//     cos(t_steady*f_env+phi)/cos(phi) >= sqrt(.5)
	//                     where t_steady*f_env+phi >= 1.2661 = acos(.3)
	//     The "most critical" situation is at the end:
	//                           t_steady*f_env+phi = acos(.3)
	//     With f_env >= 12.5 kHz (we require at least 25 kHz, the width of
	//                        one 433 MHz channel and the cos term of the
	//                        sinus addition will contain half of the
	//                        frequency) we get
	//                           t_steady*[>= 12.5 kHz] -  acos(.3) = phi 
	//                           t_steady*[>= 12.5 kHz] =  acos(.3) + phi
	//                           t_steady             = (acos(.3) + phi)/ [>= 12.5 kHz]
	//                           t_steady             < (acos(.3) + phi)/(12.5 kHz)
	//                -> phi = acos(.3) - .83ms * f_env
	//               WARNING: This gets negative for f_env > 1,450 Hz. So
	//                  when reaching envelope functions > 1,450 Hz the 
	//                  enevlope function is guarunteed alternating 
	//                  while receiving a bit and I assume the receiver
	//                  receives nothing. (PLL confusion?)
	//                  However this is not true for faster bit rates.
	//                  So I guess before calling the adding function
	//                  there must be pre-definition of the target
	//                  bitrate.
	//     Now we extract the maximum valid time from this:
	//           cos(t_steady*f_env+phi)/cos(phi) >= sqrt(.5)
	//           cos(t_steady*f_env+acos(.3) - .83ms * f_env)/cos(acos(.3) - .83ms * f_env) >= sqrt(.5)
	
	// TODO: Proceed here!
	// 0.71 ist schön und gut für p=0
	//     plot(0:.1:70, sin(0:.1:70)+sin([0:.1:70]*1.1))
	// aber für p = 1/2 pi ist das mit 0.71 nicht mehr so gut zu 
	// berechnen.
	//     plot(0:.1:70, sin(0:.1:70)+sin([0:.1:70]*1.1+pi/2))
	// Dann ist die Bit-Detektion vmtl. auch schwierig, aber das muss hier
	// irgendwie in Worte, Formeln und Code gefasst werden.
	
	
//	// Already now we can read following from the formula:
//	// * As the simulation landscape map is not that precise (using a route
//	//   obeying a 50mx50m grid without a possibility to define curves)
//	//   p4 in the first sinus term could be considered random as we are in high
//	//   frequency domain and one wavelength is c/(433.92 MHz) = 69 cm.
	
	// There are four constellations:
	// a) both streams are '0'
	// b) first stream is '1', second stream is '0'
	// c) first stream is '0', second stream is '1'
	// d) both streams are '1'
	
	// The interesting points for the simulation are:
	// * When there's destructive and constructive interference?
	//       (How fast do they alternate?)
	
	// For a) and d) keep in mind that cos(z) = cos(-z), so cos(f1-f2)
	//    is cos(f2-f1)
	
	// Following is the resulting wave disposition:
	// a) and d)
	//       f1-f2 = 1000 Hz (or f2-f1, but this doesn't matter because
	//              cos(z) = cos(-z))
	//     Pad = a3*sin(f2*t) + 2*a1*sin(f2*t+p4 +- 500Hz*t)*cos(500Hz*t+p4) 
	//     We can see a frequency difference of 500 Hz between the sinus terms
	//     which 
	// b) and c)
	//     [b) and c) results interchangable as only the interference
	//      alternation is considered. Calculating with f1 ~= f2+15 kHz]
	//     Pb = a3*sin(f2*t) + 2*a1*sin(f2*t+p4 + 15 kHz*t +- 500Hz*t)*cos(15kHz*t+p4 +- 500Hz*t)
	
	
	
}
#endif // ifndef NOCODE
;



#endif // ifndef AIR_C


