

// © 2018 Jonas Bechtel
// Licensed under the Apache License, Version 2.0; see file <prj>/meta/LICENSE


#ifndef SIM_COMMON_H
#define SIM_COMMON_H

#include <stdint.h>
#include <stdbool.h>
#include <stdio.h>

#include <math.h>

struct packet;
struct radio_conf;
struct simulator;
struct station;


uint64_t get_simtime(struct station*);

#include "assert.h"
// Thanks https://stackoverflow.com/a/10959486
// ( was https://stackoverflow.com/questions/5867834/assert-with-message )
#define assert__(x) for ( ; !(x) ; assert(x) )
// use as such: assert__(x) {printf("Assertion will fail\n");}
#define assert2(x, str) for ( ; !(x) ; assert(x) ) {printf(str == NULL ? "Assertion will fail.\n" : str);}
// use as such: assert2(x, "Program or data error.\n")

//#include "station.h"
/*
#ifndef NOCODE
#define NOCODE
#include "read_ride.c"
#undef NOCODE
#else
#include "read_ride.c"
#endif
*/


// Simulation time is at least an ~uint48_t and is counted in nanoseconds.
#define NSEC 
#define USEC *       1000ULL
#define MSEC *    1000000ULL
#define  SEC * 1000000000ULL


#define SYNC_WORD "CN" // Used by tranceiver

#define MAX_RX_SEQUENCES_PER_STATION 5

// For calcuation of dBms impact
#define EXPONENTIAL_BASIS_PWRCUM 1.5


typedef uint8_t dist;
typedef float damp;
struct S2S_channel {
	// dist: 0x00 - 0x7F: Zeitdistanz 0 - 12700 Nanosekunden
	//       0xFF = Kanal nicht aktiv
	dist dist;
	
	damp damp;
};




struct digital_packet {
	uint64_t starttime;
	uint32_t length; // how many bits(!) are to be transferred.
	uint16_t capacity; // In byte!
	uint32_t datarate; // bitps
	
	// If the number of bits is not a multiple of 8, the remainder
	// counts the bits from the LSB of the partial byte.
	// (This is a random guess how RFM69 could transmit. LSB first is easier for C.)
	// (Over SPI the bits are transferred the other way, MSB first.)
	
	// This MUST be the last member of the struct!
	uint8_t data[1]; // To be extended to one's own needs.
};

// A waveblip is a very short segment of time which contains a 
// single-frequency electromagnetical wave.
struct waveblip {
	uint32_t duration; // in ns
	
	uint32_t frequency; // TODO: make this a baseband / lowpass fr. (im Basisband)
			// frequency must leave the MSB 0 which will be used
			// in subsequent calculations (it will never be touched)
			// -> max. frequency = 2.147 GHz
	float amplitude; // in mW (a power unit! - "amplitude" may be misleading)
	uint8_t phase; // 0 = 0, 128 = pi, 255+1 = 2pi
};

static inline bool is_waveblip_invalid(struct waveblip wb) {
	return isnan(wb.amplitude);
}

// Arbritrary structure to transfer data
struct wave_sequence {
	// After this struct has gone to simulator, the seq data may not be
	// changed any more.
	// The struct will be freed by simulator.
	
	char source;
	uint8_t refcount;
	
	uint16_t capacity;
	uint32_t delay; // how long to wait till (further) transmission.
	uint16_t skip; // How many entries of seq shall be skipped.
	uint32_t segmentation; // How much time of seq shall be used. NOTE below.
	uint64_t duration; // in ns from skip to fin.
	
	// If array capacity is not reached by the data, the last seq item
	// must be followed by a waveblib whose amplitude is NaN
	struct waveblip seq[1]; // To be extended to one's own needs.
};
// NOTE concerning segmentation: when station requests segmentation, it
// gets notified when the segment has been sent. At the segmentation time
// the seq may have been altered by air.c 
// Station has to update the wave_sequence's skip variable and place a new
// send procedure.
// [It is yet unclear how the phyiscal delay between send and receive will be
// resolved. Expecting that the stations have no more distance than 3000 m.
// -> max. 10 us delay to the farest receiver station.
// So we can hopefully use the same sequence for every station before a new
// segment (typically ~ 1 ms) will be called (No need to copy wave sequences).
// But I guess one final copy will be required to resolve timings. (data
// has to be touched in every case -> phase information)]

// Following structure is recommendable:
// [Note: RX station is singular, but procedure will be executed for every
//  station.]
// 1. TX station creates wave_sequence and put it onto the event queue.
// 2. RX station has a collection of wave_sequences (many possible as 
//        some reflections also are own wave_sequences)
//        [actually the simulator has all the data store]
// 3. These wave sequences will be added, but only till on of the following
//        events happens:
//	* Some station has an earliest claim time.
//	* The segmentation of one of the addition sources is met. [TODO: This is an error then! But how to reveal that??]
// 4. The wave sequences will be split. Every RX station has an informational
//        store.


struct wave_seq_segment {
	// Is fully invalid if waves == NULL
	
	uint64_t starttime;
	
	struct S2S_channel airinfo; // THIS AFFECTS starttime! (distance delay to be added)
	
	
	// if first_element is invalid, the begin is at second element.
	struct waveblip first_element;
	// TODO: new rule: first_element _never_ can be invalid. Set instead the 
	// num_second_element +1 and copy the first element from waves into
	// first_element!
	
	int num_second_element;
	struct wave_sequence * waves;
	//int num_afterlast; // Element after the last element.
	uint32_t segment_duration; // TODO: Decide wether this is absolute or
			// min/max/... or what this variable is for.
	
};



#ifndef NAN
#error "The NAN capability of C is needed in order to mark invalidated waveblips and find them later."
#endif




void clean_wave_seq_segment (struct wave_seq_segment* wss);


/*
// TODO: Remove this?
inline uint8_t* get_ppointer(struct digital_packet* p, off_t from, size_t size) {
	if (p->length < from + size)
		return NULL;
	
	return &(p->data[from]);
};*/



// TODO: Delete radio_conf
typedef uint8_t sist;
typedef uint8_t ch;
typedef uint32_t spd;
struct radio_conf {
	ch channel;
	sist signal_strength;
	spd transmission_speed; // 1.200 kbit/s, 57600 kbit/s, ...
};




#endif // #ifndef SIM_COMMON_H


