/*****************************************************************************
 *
 * Copyright (C) 2001 Uppsala University & Ericsson AB.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors: Erik Nordstrm, <erik.nordstrom@it.uu.se>
 *
 *****************************************************************************/

#ifndef PACKET_QUEUE_C
#define PACKET_QUEUE_C

#ifndef NOCODE

#ifdef CN_PORT
#include "../defs.h"
#include "../debug.h"
#include "../routing_table.h"
#include "../aodv_timeout.h"
#include "packet_queue.h"
#define NOCODE
#include "../packetizer.c"
#include "../cli-dev.c"
#undef NOCODE



#ifdef REMOTE_CTRL
#ifdef SIMUL
#include "common/route-user.h"
#define NOCODE
#include "FrameSwitch.c"
#undef NOCODE
//#include "multihw.c"
//#include "common/app-user.h"
#else
// Expecting these two implemented by controlling application
#define MHI_ARG
bool outbound_raw    (MHI_ARG    uint8_t* data, uint8_t len, uint32_t tracking_mark);
#endif // -ifdef SIMUL -else
#endif // -ifdef REMOTE_CTRL


#else

#include "aodv-uu.h"
#endif

#define GARBAGE_COLLECT


void NS_CLASS packet_queue_init(HI_ARG_single)
{
    INIT_LIST_HEAD(&HI_USE_access PQ.head);
    HI_USE_access PQ.len = 0;

#ifdef GARBAGE_COLLECT
    /* Set up garbage collector */
    timer_init(HI_USE_arg    &HI_USE_access PQ.garbage_collect_timer, &NS_CLASS packet_queue_timeout, &HI_USE_access PQ);
	
    timer_set_timeout(HI_USE_arg   &HI_USE_access PQ.garbage_collect_timer, GARBAGE_COLLECT_TIME);
#endif
}


void NS_CLASS packet_queue_destroy(HI_ARG_single)
{
    int count = 0;
    list_t *pos, *tmp;
    
    list_foreach_safe(pos, tmp, &HI_USE_access PQ.head) {
	struct q_pkt *qp = (struct q_pkt *)pos;
	list_detach(pos);

#ifdef CN_PORT
	free(qp->p);
#else
	drop(qp->p, DROP_END_OF_SIMULATION);
#endif

	free(qp);
	count++;
	HI_USE_access PQ.len--;
    }

    debug(HI_or_NULL, LOG_INFO, 0, "Destroyed %d buffered packets!", count);
}

/* Garbage collect packets which have been queued for too long... */
int NS_CLASS packet_queue_garbage_collect(HI_ARG_single)
{
    int count = 0;
    list_t *pos, *tmp;
    struct timeval now;
    
    gettimeofday(&now, NULL);
    
    list_foreach_safe(pos, tmp, &HI_USE_access PQ.head) {
	struct q_pkt *qp = (struct q_pkt *)pos;
	if (timeval_diff(&now, &qp->q_time) > MAX_QUEUE_TIME) {
			debug(HI_or_NULL, LOG_WARNING, 0, "Removing q_pkt with len=%u.", qp->unpacked_length);
			list_detach(pos);
		
#ifdef CN_PORT
		free(qp->p);
#else
	    drop(qp->p, DROP_RTR_QTIMEOUT); 
#endif
	    
	    free(qp);
	    count++;
	    HI_USE_access PQ.len--;
	}
    }
    
    if (count) {
        DEBUG(LOG_DEBUG, 0, "Removed %d packet(s)!", count);
    }
    
    return count;
}


/* Buffer a packet in a FIFO queue. Implemented as a linked list,
   where we add elements at the end and remove at the beginning.... */

void NS_CLASS packet_queue_add(HI_ARG    Packet * p, struct in_addr dest_addr, uint16_t len)
{
    struct q_pkt *qp;
#ifndef CN_PORT
    struct hdr_ip *ih;
    struct hdr_cmn *ch = HDR_CMN(p);
    
    ih = HDR_IP(p);
    
    assert((unsigned int)ih->daddr() == dest_addr.s_addr);
#endif
	

	if (HI_USE_access PQ.len >= MAX_QUEUE_LENGTH) {
		debug(HI_or_NULL, LOG_DEBUG, 0, "MAX Queue length! Removing first packet.");
		if (!list_empty(&HI_USE_access PQ.head)) {
			qp = (struct q_pkt *)HI_USE_access PQ.head.next;
	    
			list_detach(HI_USE_access PQ.head.next);
#ifdef CN_PORT
			free(qp->p);
#else
			drop(qp->p, DROP_RTR_QFULL);
#endif
			free(qp);
			HI_USE_access PQ.len--;
	}
    }

    qp = (struct q_pkt *) malloc(sizeof(struct q_pkt));
    
    if (qp == NULL) {
	fprintf(stderr, "Malloc failed!\n");
	exit(-1);
    }

    qp->p = p;

    qp->dest_addr = dest_addr;
	qp->unpacked_length = len;
    gettimeofday(&qp->q_time, NULL);
    
	list_add_tail(&HI_USE_access PQ.head, &qp->l);

	HI_USE_access PQ.len++;

#ifdef CN_PORT
	debug(hi, LOG_INFO, 0, "buffered pkt to %s qlen=%u",
	  ip_to_str(dest_addr), hi->PQ.len);
#else
	debug(HI_or_NULL, LOG_INFO, 0, "buffered pkt to %s uid=%d qlen=%u",
	  ip_to_str(dest_addr), ch->uid(), HI_USE_access PQ.len);
#endif
}

#endif // -ifndef NOCODE
#ifdef CN_PORT
void NS_CLASS packet_queue_add_short(struct host_info* hi, struct unpacked_L2_header* ul2h, uint8_t dst, uint16_t unpacked_length)
#ifndef NOCODE
{
	struct in_addr dest_addr;
	set_addr_byte(&dest_addr, 0, 192);
	set_addr_byte(&dest_addr, 1, 168);
	set_addr_byte(&dest_addr, 2, 1);
	set_addr_byte(&dest_addr, 3, dst);
	
	packet_queue_add(hi, ul2h, dest_addr, unpacked_length);
}
#endif // -ifndef NOCODE
;

#endif // -ifdef CN_PORT

#ifndef NOCODE

int NS_CLASS packet_queue_set_verdict(HI_ARG   struct in_addr dest_addr, int verdict)
{
    int count = 0;
    rt_table_t *rt, *next_hop_rt, *inet_rt = NULL;
    list_t *pos, *tmp;

    double delay = 0;
#define ARP_DELAY 0.005
    
    if (verdict == PQ_ENC_SEND) {
	inet_rt = rt_table_find(HI_or_NULL, dest_addr);
	
	if (!inet_rt)
	    return -1;
	rt = rt_table_find(HI_or_NULL, inet_rt->next_hop);
    } else {
	rt = rt_table_find(HI_or_NULL, dest_addr);
    }
    
	bool break_loop = false;
	// we use backward version as otherwise an erroneus distribution procedure
	// could put the packet back on list leading to an endless loop.
	list_foreach_backward_safe(pos, tmp, &HI_USE_access PQ.head) {
		struct q_pkt *qp = (struct q_pkt *)pos;
	
		if (qp->dest_addr.s_addr == dest_addr.s_addr) {
			
			
			switch (verdict) {
			case PQ_SEND:
				
				if (!rt)
					return -1;
				
				bool res = 
#if (defined REMOTE_CTRL) && (defined SIMUL)
				hi->multihw_instance->app->rti->outbound_raw(
					hi->multihw_instance->app->rti,NULL,qp->unpacked_length,0);
#elif (defined REMOTE_CTRL) && (defined BEAGLE)
				outbound_raw(NULL, qp->unpacked_length, 0);
#endif
				if (!res) {
					DEBUG(LOG_DEBUG, 0, "Not sending of more packets as buffer is full.");
					break_loop = true;
					break;
				}
					DEBUG(LOG_DEBUG, 0, "Buffer has space.");
				
				list_detach(pos);
				
		/* Apparently, the link layer implementation can't handle
		 * a burst of packets. So to keep ARP happy, buffered
		 * packets are sent with ARP_DELAY seconds between
		 * sends. */
#ifdef CN_PORT
				
				
				process_packet_short(hi, qp->p, qp->unpacked_length, PROCESS_OUTBOUND, 0);
				free(qp->p);
#else
			sendPacket(qp->p, rt->next_hop, delay);
#endif
				delay += ARP_DELAY;
				break;
			case PQ_DROP:
				list_detach(pos);
#ifdef CN_PORT
				free(qp->p);
#else
				drop(qp->p, DROP_RTR_NO_ROUTE);
#endif
				break;
#ifdef CONFIG_GATEWAY
			case PQ_ENC_SEND:
				list_detach(pos);
				
				if (!rt)
					return -1;
				
				DEBUG(LOG_DEBUG, 0, "Encap. PQ_ENC_SEND");
				
				qp->p = pkt_encapsulate(qp->p, inet_rt->next_hop);
				
				if (!qp->p) {
					DEBUG(LOG_ERR, 0, "Encapsulation failed...");
				}
				
				sendPacket(qp->p, rt->next_hop, delay);
				delay += ARP_DELAY;
				break;
#endif /* CONFIG_GATEWAY */
				
			} // switch
			
			if (break_loop)
				break;
			
			free(qp);
			count++;
			HI_USE_access PQ.len--;
		}
	}

    /* Update rt timeouts */
    if (rt && rt->state == VALID && 
	(verdict == PQ_SEND || verdict == PQ_ENC_SEND)) {

		if (dest_addr.s_addr
				!= DEV_IFINDEX(HI_or_NULL, rt->ifindex).ipaddr.s_addr) {

	    if (verdict == PQ_ENC_SEND && inet_rt)
		rt_table_update_timeout(HI_USE_arg   inet_rt, HI_USE_access active_route_timeout);
	    
	    rt_table_update_timeout(HI_USE_arg   rt, HI_USE_access active_route_timeout);
	    
	    next_hop_rt = rt_table_find(HI_or_NULL, rt->next_hop);
	    
	    if (next_hop_rt && next_hop_rt->state == VALID && 
		next_hop_rt->dest_addr.s_addr != rt->dest_addr.s_addr)
		rt_table_update_timeout(HI_or_NULL, next_hop_rt, HI_USE_access active_route_timeout);
	}
	
	DEBUG(LOG_INFO, 0, "SENT %d packets to %s qlen=%u",
	      count, ip_to_str(dest_addr), HI_USE_access PQ.len);
    } else if (verdict == PQ_DROP) {
	DEBUG(LOG_INFO, 0, "DROPPED %d packets for %s!",
	      count, ip_to_str(dest_addr));
    }

    return count;
}

#endif // -ifndef NOCODE

#endif // -ifndef PACKET_QUEUE_C

