

// © 2018 Jonas Bechtel
// Licensed under the Apache License, Version 2.0; see file <prj>/meta/LICENSE



#ifndef RFM69_REG_H
#define RFM69_REG_H


// Besides: We are not the first ones to define these registers.
//  But we do it in a nice manner!

// https://github.com/cristi85/RFM69
// https://github.com/LowPowerLab/RFM69/blob/master/RFM69registers.h
// For Arduino: https://codebender.cc/library/RFM69#RFM69registers.h
// https://github.com/nayem-cosmic/RFM69-Library-AVR/
// Broad: http://www.airspayce.com/mikem/arduino/RadioHead/

// Sources: google, http://www.avrfreaks.net/forum/rfm69-avr-gcc-library, https://community.atmel.com/comment/2140001


// Startup and preamble timing

// From Doc:

// Fig. 17 - TS_RE = Rx Startup - No AGC, no AFC
// Tana   - Analog FE's group delay
// Tcf    - Channel Filter's group delay
// Tdcc   - DC Cutoff's group delay
// <-> Received Packe Preamble may start
// Trssi  - RSSI sampling
// Trssi  - RSSI sampling

// Fig. 18 - TS_RE_AGC = Rx Startup - AGC, no AFC
// Tana   - Analog FE's group delay
// Tcf    - Channel Filter's group delay
// Tdcc   - DC Cutoff's group delay
// <-> Received Packe Preamble may start
// Trssi  - RSSI sampling
// Trssi  - RSSI sampling
// Tcf    - Channel Filter's group delay
// Tdcc   - DC Cutoff's group delay
// Trssi  - RSSI sampling

// Fig. 19 - TS_RE_AGC = Rx Startup - AGC and AFC
// Tana   - Analog FE's group delay
// Tcf    - Channel Filter's group delay
// Tdcc   - DC Cutoff's group delay
// <-> Received Packe Preamble may start
// Trssi  - RSSI sampling
// Trssi  - RSSI sampling
// Tcf    - Channel Filter's group delay
// Tdcc   - DC Cutoff's group delay
// Trssi  - RSSI sampling
// Tafc   - AFC
// Tpllafc- PLL lock
// Tcf    - Channel Filter's group delay
// Tdcc   - DC Cutoff's group delay

// The different timings shown above are as follows:
// * Group delay of the analog front end: Tana = 20 us
// * Channel filter’s group delay in FSK mode: Tcf = 21 / (4.RxBw)
// * Channel filter’s group delay in OOK mode: Tcf = 34 / (4.RxBw)
// * DC Cutoff’s group delay: Tdcc = max(8 , 2^(round(log2(8.RxBw.Tbit)+1)) / (4.RxBw)
// * PLL lock time after AFC adjustment: Tpllafc = 5 / PLLBW (PLLBW = 300 kHz)
// * AFC sample time: Tafc = 4 x Tbit (also denoted TS_AFC in the general specification)
// * RSSI sample time: Trssi = 2 x int(4.RxBw.Tbit)/(4.RxBw) (aka TS_RSSI = 2*Tbit)
// Note: The above timings represent maximum settling times, and shorter settling times may be observed in real cases



// TODO: This can be public:
#define FXOSC 32000000 // 32 MHz. Datenblatt, S. 13
#define FSTEP 61.03515625 // FXOSC/2^19. Datenblatt, S. 13



// Regfile definitions


// Datenblatt, S. 63

#define FIFO_SIZE 66

#define RegFifo 0x00
#define cbFifo combine_regdata(RegFifo, Fifo, REGFILE_RW | REGFILE_WRITE_SIMPLE, 0xff, 0)

#define RegOpMode 0x01

#define RFM_MODE_SLEEP 0   // in RegOpMode: Mode = 0 0 0 (ListenOn = 0)
#define RFM_MODE_STANDBY 1 // in RegOpMode: Mode = 0 0 1 (ListenOn = 0)
#define RFM_MODE_FS 2      // in RegOpMode: Mode = 0 1 0 (ListenOn = 0)
#define RFM_MODE_TX 3      // in RegOpMode: Mode = 0 1 1 (ListenOn = 0)
#define RFM_MODE_RX 4      // in RegOpMode: Mode = 1 0 0 (ListenOn = 0)


// pure version: only sets RegOpMode:Mode
#define cbMode combine_regdata(RegOpMode, Mode, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 0x7, 2)
#define cbListenAbort combine_regdata(RegOpMode, ListenAbort, REGFILE_WRITE, 0x1, 5)
#define cbListenOn combine_regdata(RegOpMode, ListenOn, REGFILE_RW | REGFILE_NEEDS_READ, 0x1, 6)
#define cbSequencerOff combine_regdata(RegOpMode, SequencerOff, REGFILE_RW | REGFILE_NEEDS_READ, 0x1, 7)




#define RegDataModul 0x02

#define cbDataMode combine_regdata(RegDataModul, DataMode, REGFILE_RW, 0x3, 5)
#define PACKET_MODE 0
#define CONTINUOUS_MODE 2
#define CONTINUOUS_MODE_WO_BIT_SYNC 3

#define cbModulationType combine_regdata(RegDataModul, ModulationType, REGFILE_RW, 0x03, 3)
#define MODULATION_FSK 0
#define MODULATION_OOK 1
#define cbModulationShaping combine_regdata(RegDataModul, ModulationShaping, REGFILE_RW, 0x3, 0)
#define NO_SHAPING 0 // FSK, OOK
#define GAUSS_BT1_0 1 // FSK
#define GAUSS_BT0_5 2 // FSK
#define GAUSS_BT0_3 3 // FSK
#define CUTOFF_BR 1 // OOK
#define CUTOFF_2_BR 2 // OOK


#define RegBitrateMsb 0x03
// Datenblatt, S. 64
#define RegBitrateLsb 0x04

#define cbBitrateMsb combine_regdata(RegBitrateMsb, BitrateMsb, REGFILE_RW | REGFILE_WRITE_SIMPLE, 0xff, 0)
#define cbBitrateLsb combine_regdata(RegBitrateLsb, BitrateLsb, REGFILE_RW | REGFILE_WRITE_SIMPLE, 0xff, 0)


#define RegFdevMsb    0x05
#define RegFdevLsb    0x06

#define cbFdevMsb combine_regdata(RegFdevMsb, FdevMsb, REGFILE_RW | REGFILE_WRITE_SIMPLE, 0x1f, 0)
#define cbFdevLsb combine_regdata(RegFdevLsb, FdevLsb, REGFILE_RW | REGFILE_WRITE_SIMPLE, 0xff, 0)
// Frequency deviation F_dev = Fstep * Fdev(15:0)  [die meinen Fdev(13:0) ?]
// F_dev + Bitrate/2 <= 500 kHz!

#define RegFrfMsb     0x07
#define RegFrfMid     0x08
#define RegFrfLsb     0x09

#define cbFrfMsb combine_regdata(RegFrfMsb, FrfMsb, REGFILE_RW | REGFILE_WRITE_SIMPLE, 0xff, 0)
#define cbFrfMid combine_regdata(RegFrfMid, FrfMid, REGFILE_RW | REGFILE_WRITE_SIMPLE, 0xff, 0)
#define cbFrfLsb combine_regdata(RegFrfLsb, FrfLsb, REGFILE_RW | REGFILE_WRITE_SIMPLE, 0xff, 0)
// Carrier frequency F_RF = Fstep * Frf(23:0)


// Datenblatt, S. 66

#define RegPaLevel     0x11
#define cbPa0On       combine_regdata(RegPaLevel, Pa0On,     REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 0x01, 7)
#define cbPa1On       combine_regdata(RegPaLevel, Pa1On,     REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 0x01, 6)
#define cbPa2On       combine_regdata(RegPaLevel, Pa2On,     REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 0x01, 5)
#define cbOutputPower combine_regdata(RegPaLevel, OutputPwr, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 0x1f, 0)


#define RegOcp         0x13
#define cbOcpOn combine_regdata(RegOcp, OcpOn, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 1, 4)
#define cbOcpTrim combine_regdata(RegOcp, OcpTrim, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 0xf, 0)
// double combined:
#define cbRegOcp combine_regdata(RegOcp, ROcp, REGFILE_RW | REGFILE_WRITE_SIMPLE, 0x1f, 0)


// Datenblatt, S. 67

#define RegLna         0x18
#define cbLnaZin combine_regdata(RegLna, LnaZin, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 1, 7)
#define cbLnaCurrentGain combine_regdata(RegLna, LnaCurrentGain, REGFILE_READ, 7, 3)
#define cbLnaGainSelect  combine_regdata(RegLna, LnaGainSelect, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 7, 0)

#define RegRxBw        0x19
#define cbDocFreq  combine_regdata(RegRxBw, DocFreq,  REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 7, 5)
#define cbRxBwMant combine_regdata(RegRxBw, RxBwMant, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 3, 3)
#define cbRxBwExp  combine_regdata(RegRxBw, RxBwExp,  REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 7, 0)


// Datenblatt, S. 68

#define RegAfcFei      0x1E
#define cbFeiDone  combine_regdata(RegAfcFei, FeiDone,  REGFILE_READ, 1, 6)
#define cbFeiStart combine_regdata(RegAfcFei, FeiStart, REGFILE_WRITE | REGFILE_WRITE_NEEDS_READ, 1, 5)
#define cbAfcDone  combine_regdata(RegAfcFei, AfcDone,  REGFILE_READ, 1, 4)
#define cbAfcAutoclearOn combine_regdata(RegAfcFei, AfcAutoclearOn, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 1, 3)
#define cbAfcAutoOn combine_regdata(RegAfcFei, AfcAutoOn, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 1, 2)
#define cbAfcClear  combine_regdata(RegAfcFei, AfcClear,  REGFILE_WRITE | REGFILE_WRITE_NEEDS_READ, 1, 1)
#define cbAfcStart  combine_regdata(RegAfcFei, AfcStart,  REGFILE_WRITE | REGFILE_WRITE_NEEDS_READ, 1, 0)
#define RegAfcMsb      0x1F
#define cbAfcValueMsb      combine_regdata(RegAfcMsb, AfcValueMsb, REGFILE_READ, 0xff, 0)
#define RegAfcLsb      0x20
#define cbAfcValueLsb      combine_regdata(RegAfcLsb, AfcValueLsb, REGFILE_READ, 0xff, 0)
#define RegFeiMsb      0x21
#define cbFeiValueMsb combine_regdata(RegFeiMsb, FeiValueMsb, REGFILE_READ, 0xff, 0)
#define RegFeiLsb      0x22
#define cbFeiValueLsb combine_regdata(RegFeiLsb, FeiValueLsb, REGFILE_READ, 0xff, 0)

#define RegRssiConfig  0x23
#define cbRssiDone combine_regdata(RegRssiConfig, RssiDone, REGFILE_READ, 1, 1)
#define cbRssiStart combine_regdata(RegRssiConfig, RssiStart, REGFILE_WRITE | REGFILE_WRITE_SIMPLE, 1, 0)

#define RegRssiValue   0x24
#define cbRssiValue combine_regdata(RegRssiValue, RssiValue, REGFILE_READ, 0xff, 0)


// Datenblatt, S. 69

#define RegDioMapping1 0x25
#define cbDio0Mapping  combine_regdata(RegDioMapping1, Dio0Mapping, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 3, 6)
#define cbDio1Mapping  combine_regdata(RegDioMapping1, Dio1Mapping, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 3, 4)
#define cbDio2Mapping  combine_regdata(RegDioMapping1, Dio2Mapping, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 3, 2)
#define cbDio3Mapping  combine_regdata(RegDioMapping1, Dio3Mapping, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 3, 0)

#define RegDioMapping2 0x26
#define cbDio4Mapping  combine_regdata(RegDioMapping2, Dio4Mapping, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 3, 6)
#define cbDio5Mapping  combine_regdata(RegDioMapping2, Dio5Mapping, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 3, 4)
#define cbClkOut       combine_regdata(RegDioMapping2, ClkOut,      REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 7, 0)

// For the DIO mappings
#define CLK_FXOSC       0
#define CLK_FXOSC_div2  1
#define CLK_FXOSC_div4  2
#define CLK_FXOSC_div8  3
#define CLK_FXOSC_div16 4
#define CLK_FXOSC_div32 5
#define CLK_RC          6
#define CLK_OFF         7
// All those are ONLY VALID FOR PACKET MODE!!!
#define ClkOut5          0
#define DataOut5         1
#define ModeReady5       3
#define Timeout4         0
#define Rssi_TxReady_4 1
#define Rssi4    Rssi_TxReady_4
#define TxReady4 Rssi_TxReady_4
#define RxReady4         2
#define PllLock4         3
// All those are ONLY VALID FOR PACKET MODE!!!
#define FifoFull3        0
#define Rssi_TxReady_3 Rssi_TxReady_4
#define Rssi3 Rssi4
#define TxReady3 TxReady4
#define SyncAddress3     2
#define PllLock3 PllLock4
#define FifoNotEmpty2    0
#define Data2            1
#define AutoMode2        3
// All those are ONLY VALID FOR PACKET MODE!!!
#define FifoLevel1       0
#define FifoFull1        1
#define FifoNotEmpty1     2
#define PllLock1 PllLock4
#define CrcOk_PacketSent_0 0
#define Crc0 CrcOk_PacketSent_0
#define PacketSent0 CrcOk_PacketSent_0
#define PayloadReady_TxReady_0 1
#define PayloadReady0 PayloadReady_TxReady_0
#define TxReady0 PayloadReady_TxReady_0
#define SyncAddress0     2
#define Rssi_PllLock_0   4
#define Rssi0 Rssi_PllLock_0
#define PllLock0 Rssi_PllLock_0






#define RegIrqFlags1   0x27
#define cbModeReady combine_regdata(RegIrqFlags1, ModeReady, REGFILE_READ, 1, 7)
#define cbRxReady   combine_regdata(RegIrqFlags1, RxReady,   REGFILE_READ, 1, 6)
#define cbTxReady   combine_regdata(RegIrqFlags1, TxReady,   REGFILE_READ, 1, 5)
#define cbPllLock   combine_regdata(RegIrqFlags1, PllLock,   REGFILE_READ, 1, 4)
#define cbRssi      combine_regdata(RegIrqFlags1, Rssi,      REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 1, 3) // Don't know why Rssi could be writable?
#define cbTimeout   combine_regdata(RegIrqFlags1, Timeout,   REGFILE_READ, 1, 2)
#define cbAutoMode  combine_regdata(RegIrqFlags1, AutoMode,  REGFILE_READ, 1, 1)
#define cbSyncAddressMatch combine_regdata(RegIrqFlags1, SAM, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 1, 0)


// Datenblatt, S. 70
#define RegIrqFlags2   0x28
#define cbFifoFull     combine_regdata(RegIrqFlags2, FifoFull,  REGFILE_READ, 1, 7)
#define cbFifoNotEmpty combine_regdata(RegIrqFlags2, FifoNE,    REGFILE_READ, 1, 6)
#define cbFifoLevel    combine_regdata(RegIrqFlags2, FifoLevel, REGFILE_READ, 1, 5)
// WRITE_SIMPLE may be dangerous?
#define cbFifoOverrun  combine_regdata(RegIrqFlags2, FifoOR,    REGFILE_RW |REGFILE_WRITE_SIMPLE, 1, 4)
#define cbPacketSent   combine_regdata(RegIrqFlags2, PacketSent,REGFILE_READ, 1, 3)
#define cbPayloadReady combine_regdata(RegIrqFlags2, PayloadRdy,REGFILE_READ, 1, 2)
#define cbCrcOK        combine_regdata(RegIrqFlags2, CrcOk,     REGFILE_READ, 1, 1)

#define RegRssiThresh   0x29
#define cbRssiThreshold combine_regdata(RegRssiThresh, RssiThresh, REGFILE_RW | REGFILE_WRITE_SIMPLE, 0xff, 0)


// Datenblatt, S. 71
#define RegSyncConfig   0x2e
#define cbSyncOn        combine_regdata(RegSyncConfig, SyncOn,  REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 1, 7)
#define cbFifoFillCondition combine_regdata(RegSyncConfig, FiFiCo, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 1, 6)
#define cbSyncSize      combine_regdata(RegSyncConfig, SyncSize,REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 0x7, 3)
#define cbSyncTol       combine_regdata(RegSyncConfig, SyncTol, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 0x7, 0)

#define RegSyncValue    0x2f // bis 0x36
#define cbSyncValue1    combine_regdata(RegSyncValue  , SyncV1, REGFILE_RW | REGFILE_WRITE_SIMPLE, 0xff, 0)
#define cbSyncValue2    combine_regdata(RegSyncValue+1, SyncV1, REGFILE_RW | REGFILE_WRITE_SIMPLE, 0xff, 0)
#define cbSyncValue3    combine_regdata(RegSyncValue+2, SyncV1, REGFILE_RW | REGFILE_WRITE_SIMPLE, 0xff, 0)
#define cbSyncValue4    combine_regdata(RegSyncValue+3, SyncV1, REGFILE_RW | REGFILE_WRITE_SIMPLE, 0xff, 0)
#define cbSyncValue5    combine_regdata(RegSyncValue+4, SyncV1, REGFILE_RW | REGFILE_WRITE_SIMPLE, 0xff, 0)
#define cbSyncValue6    combine_regdata(RegSyncValue+5, SyncV1, REGFILE_RW | REGFILE_WRITE_SIMPLE, 0xff, 0)
#define cbSyncValue7    combine_regdata(RegSyncValue+6, SyncV1, REGFILE_RW | REGFILE_WRITE_SIMPLE, 0xff, 0)
#define cbSyncValue8    combine_regdata(RegSyncValue+7, SyncV1, REGFILE_RW | REGFILE_SIMPLE_WRITE, 0xff, 0)



// Datenblatt, S. 72
#define RegPacketConfig1 0x37

#define PACKET_FORMAT_FIXEDlength 0
#define PACKET_FORMAT_VARIABLElength 1
#define cbPacketFormat combine_regdata(RegPacketConfig1, PacketFormat, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 1, 7)

#define DCFREE_none 0
#define DCFREE_Manchester 1
#define DCFREE_Whitening 2
#define cbDcFree       combine_regdata(RegPacketConfig1, DcFree,       REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 3, 5)

#define cbCrcOn        combine_regdata(RegPacketConfig1, CrcOn,        REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 1, 4)
#define cbCrcAutoClearOff combine_regdata(RegPacketConfig1, CrcACO,    REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 1, 3)

#define ADDRESS_FILTER_none 0
#define ADDRESS_FILTER_NODE 1
#define ADDRESS_FILTER_NODE_OR_BROADCAST 2
#define cbAddressFiltering combine_regdata(RegPacketConfig1, AddressF, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 3, 1)

#define RegPayloadLength 0x38
#define cbPayloadLength  combine_regdata(RegPayloadLength, PayL, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 0xff, 0)


// Datenblatt, S. 73
#define RegFifoThresh  0x3C

#define COND_FifoThreshold 0
#define COND_FifoNotEmpty 1
#define cbTxStartCondition combine_regdata(RegFifoThresh, TxSC,    REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 1, 7)

#define cbFifoThreshold combine_regdata(RegFifoThresh, FifoThresh, REGFILE_RW | REGFILE_WRITE_NEEDS_READ, 0x7f, 0)



// Datenblatt, S. 74

#define RegTestLna     0x58
#define cbSensitivityBoost combine_regdata(RegTestLna, SensitivityBoost, REGFILE_RW | REGFILE_WRITE_SIMPLE, 0xff, 0)

#define RegTestPa1     0x5A
#define cbPa20dBm1 combine_regdata(RegTestPa1, Pa20dBm1, REGFILE_RW | REGFILE_WRITE_SIMPLE, 0xff, 0)
#define cbRegTestPa1 cbPa20dBm1 // auxiliary

#define RegTestPa2     0x5C
#define cbPa20dBm2 combine_regdata(RegTestPa2, Pa20dBm2, REGFILE_RW | REGFILE_WRITE_SIMPLE, 0xff, 0)
#define cbRegTestPa2 cbPa20dBm2 // auxiliary









#define combine_regdata(name, variable, access, bitmask, bitshift) \
				(((uint32_t)(name) << (uint32_t)20) + ((uint32_t)(access) << (uint32_t)16) + ((uint32_t)(bitmask) << (uint32_t)8) + (uint32_t)(bitshift))
		// (access = 0x4 READ | 0x2 WRITE)
		// (bitshift = 4 for bits 4,5,6 so they will be at 0,1,2)
		// (bitmask = 0x7 for bits 4,5,6 which appear as 0,1,2)

#define REGFILE_READ 4
#define REGFILE_WRITE 2
#define REGFILE_RW (REGFILE_READ|REGFILE_WRITE) // auxiliary

#define REGFILE_WRITE_NEEDS_READ 8
#define REGFILE_WRITE_SIMPLE 0


#define regfile_is_read(cb) ((((cb) >> 16) & 0x0f & REGFILE_READ)/REGFILE_READ)
#define regfile_is_write(cb) ((((cb) >> 16) & 0x0f & REGFILE_WRITE)/REGFILE_WRITE)
#define regfile_write_needs_read(cb) ((((cb) >> 16) & 0x0f & REGFILE_WRITE_NEEDS_READ)/REGFILE_WRITE)
#define regfile_name(cb) (((cb)>>20) & 0xfff)
#define regfile_bitmask(cb) (((cb) >> 8) & 0xff)
#define regfile_bitshift(cb) ((cb) & 0xff)





#endif // ifndef RFM69_REG_H

